# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import re
import os
import fnmatch
import stat
import codecs
import shutil
import hashlib
from collections import deque
from distutils import dir_util, file_util
#from zipfile import ZipFile
import zipfile
import tarfile
import site
import logging
import tempfile
import itertools
import glob
import datetime
import re
import string
import random

try:
    from pathlib import Path as Pathlib
except ImportError:
    Pathlib = None

try:
    import cPickle as pickle
except ImportError:
    import pickle

from .compat import *
from . import environ
from .string import String, ByteString


logger = logging.getLogger(__name__)


class Path(String):
    """Provides a more or less compatible interface to the pathlib.Path api of
    python 3 that can be used across py2 and py3 codebases. However, The Path
    instances are actually string instances and are always resolved

    This finally brings into a DRY location my path code from testdata, stockton,
    and bang, among others. I've needed this functionality. I've also tried to standardize
    the interface to be very similar to Pathlib so you can, hopefully, swap between
    them

    Directories: 
        * path: /parent/basename
        * fileroot: basename
        * basename: basename
        * filename: basename
        * directory: /parent
        * parent: /parent
        * stempath: /parent/basename
        * ext: ""
        * suffix: ""

    Files: 
        * path: /parent/fileroot.ext
        * fileroot: fileroot
        * filename: fileroot.ext
        * basename: fileroot.ext
        * directory: /parent
        * parent: /parent
        * stempath: /parent/fileroot
        * ext: ext
        * suffix: .ext

    https://en.wikipedia.org/wiki/Fully_qualified_name#Filenames_and_paths
    https://docs.python.org/3/library/pathlib.html
    https://github.com/python/cpython/blob/3.8/Lib/pathlib.py
    https://docs.python.org/3/library/os.path.html

    Path takes advantage of these semi-standard python modules:
        * https://docs.python.org/3/library/shutil.html
        * https://github.com/python/cpython/blob/2.7/Lib/distutils/dir_util.py)
        * https://github.com/python/cpython/blob/2.7/Lib/distutils/file_util.py
    """
    @property
    def permissions(self):
        # https://stomp.colorado.edu/blog/blog/2010/10/22/on-python-stat-octal-and-file-system-permissions/
        mode = stat.S_IMODE(os.stat(self.path).st_mode)
        mode = oct(mode)
        return mode.replace("o", "")

    @permissions.setter
    def permissions(self, v):
        self.chmod(v)

    @property
    def pathlib(self):
        if is_py2:
            raise NotImplementedError()
        return Pathlib(self.path)

    @property
    def parts(self):
        """A tuple giving access to the path’s various components

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.parts
        """
        if is_py2:
            parts = self.path.split("/")
            if not parts[0]:
                parts[0] = self.anchor

        else:
            parts = self.pathlib.parts

        return parts

    @property
    def root(self):
        """A string representing the (local or global) root, if any

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.root
        """
        if is_py2:
            ret, tail = os.path.splitdrive(self.path)
            if not ret:
                ret = "/"

        else:
            ret = self.pathlib.root

        return ret

    @property
    def anchor(self):
        """The concatenation of the drive and root

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.anchor
        """
        if is_py2:
            ret = self.root
            if ":" in ret:
                ret = "{}\\".format(ret)

        else:
            ret = self.pathlib.anchor

        return ret

    @property
    def parents(self):
        """An immutable sequence providing access to the logical ancestors of the path

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.parents
        """
        if is_py2:
            parents = []
            parent = self
            while True:
                p = parent.parent
                if p != parent:
                    parents.append(p)
                    parent = p
                else:
                    break

        else:
            parents = [self.create_dir(p) for p in self.pathlib.parents]

        return parents

    @property
    def parent(self):
        """The logical parent of the path

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.parent
        """
        return self.create_dir(os.path.dirname(self.path))

    @property
    def directory(self):
        """return the directory portion of a directory/fileroot.ext path"""
        return self.parent

    @property
    def dirname(self):
        return self.parent

    @property
    def basedir(self):
        return self.parent

    @property
    def name(self):
        """A string representing the final path component, excluding the drive and root

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.name
        """
        return self.basename

    @property
    def filename(self):
        return self.basename

    @property
    def basename(self):
        """Return the fileroot.ext portion of a directory/fileroot.ext path"""
        return os.path.basename(self.path)

    @property
    def fileroot(self):
        """Return the fileroot portion of a directory/fileroot.ext path"""
        # https://stackoverflow.com/questions/2235173/
        # https://stackoverflow.com/a/2235762/5006
        fileroot, ext = os.path.splitext(self.name)
        return fileroot

    @property
    def stem(self):
        """The final path component, without its suffix

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.stem
        """
        return self.fileroot

    @property
    def stempath(self):
        """The complete path, without its suffix

        :Example:
            p = Path("/foo/bar/fileroot.ext")
            p.stempath # /foo/bar/fileroot

        :returns: String, the full path and fileroot without suffix (ext)
        """
        return os.path.splitext(self.path)[0]

    @property
    def suffix(self):
        """The file extension of the final component

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.suffix
        """
        fileroot, ext = os.path.splitext(self.name)
        return ext

    @property
    def ext(self):
        return self.suffix.lstrip(".")

    @property
    def extension(self):
        return self.ext

    @property
    def suffixes(self):
        """A list of the path’s file extensions

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.suffixes
        """
        if is_py2:
            suffixes = ["." + s for s in self.basename.split(".")[1:]]

        else:
            suffixes = self.pathlib.suffixes

        return suffixes

    @classmethod
    def path_class(cls):
        """Return the Path class this class will use, this is a method because 
        we couldn't make them all Path class properties because they are defined
        after Path"""
        return Path

    @classmethod
    def dir_class(cls):
        """Return the Dirpath class to use"""
        return Dirpath

    @classmethod
    def file_class(cls):
        """Return the Filepath class to use"""
        return Filepath

    @classmethod
    def create_file(cls, *parts, **kwargs):
        kwargs["path_class"] = cls.file_class()
        return cls.create(*parts, **kwargs)
        #return kwargs["path_class"](*parts, **kwargs)

    @classmethod
    def create_dir(cls, *parts, **kwargs):
        kwargs["path_class"] = cls.dir_class()
        return cls.create(*parts, **kwargs)
        #return kwargs["path_class"](*parts, **kwargs)

    @classmethod
    def create_path(cls, *parts, **kwargs):
        kwargs["path_class"] = cls.path_class()
        return cls.create(*parts, **kwargs)
        #return kwargs["path_class"](*parts, **kwargs)

    @classmethod
    def create(cls, *parts, **kwargs):
        """Create a path instance using the full inferrencing (guessing code) of
        cls.path_class().__new__()"""
        # we want inference to work so we don't want any path_class being passed
        # to the __new__ method but we will use it to create the instance if
        # passed in
        path_class = kwargs.pop("path_class", cls.path_class())
        return path_class(*parts, **kwargs)

    @classmethod
    def create_as(cls, instance, path_class, **kwargs):
        """Used by .__new__() to convert a Path to one of the children. This is 
        a separate method so it could be augmented by children classes if desired

        take note that this has a different signature than all the other create_*
        methods

        :param instance: Path, a Path instance created by __new__()
        :param path_class: type, the path class that was passed in to __new__()
        :param **kwargs: the keywords passed into __new__()
        :returns: Path, either the same instance or a different one
        """
        instance.path = kwargs["path"]

        if path_class or (cls is not cls.path_class()):
            # makes sure if you've passed in any class explicitely, even the Path
            # class, then don't try and infer anything
            pass

        else:
            if instance.is_file():
                instance = instance.as_file()

            elif instance.is_dir():
                instance = instance.as_dir()

            else:
                # let's assume a file if it has an extension
                if instance.ext:
                    instance = instance.as_file()

        if kwargs.get("touch", kwargs.get("create", False)):
            instance.touch()

        return instance

    @classmethod
    def splitparts(cls, *parts, **kwargs):
        """Does the opposite of .join()


        :param *parts: mixed, as many parts as you pass in as arguments
        :param **kwargs:
            regex - the regex used to split the parts up
            root - if a root is needed this will be used
        :returns: list, all the normalized parts
        """
        ps = []
        regex = kwargs.get("regex", r"[\\/]+")
        root = kwargs.get("root", "/")
        path_class = cls.path_class()

        for p in parts:
            if isinstance(p, (basestring, path_class)) or not isinstance(p, Iterable):
                s = String(p)

                #if s and p is not None: # if you want to filter None
                if s:
                    for index, pb in enumerate(re.split(regex, s)):
                    #for index, pb in enumerate(s.split("/")):
                        if pb:
                            ps.append(re.sub(regex, "", pb))

                        else:
                            if root and not ps and index == 0:
                                ps.append(root)

                else:
                    if not ps and root:
                        ps.append(root)

            else:
                for pb in cls.splitparts(*p, **kwargs):
                    if root and pb == root:
                        if not ps:
                            ps.append(pb)

                    else:
                        ps.append(pb)

        return ps

    @classmethod
    def joinparts(cls, *parts, **kwargs):
        """like os.path.join but normalizes for directory separators

        Differences from the standard os.path.join:

            >>> os.path.join("/foo", "/bar/", "/che")
            '/che'

            >>> Path.join("/foo", "/bar/", "/che")
            '/foo/bar/che'

            >>> Path.join("", "bar")
            '/bar'

        :param *parts: mixed, as many parts as you pass in as arguments
        :param **kwargs:
            sep - if you want to join with a different separator than os.sep
        :returns: str, all the parts joined together
        """
        ps = cls.splitparts(*parts, **kwargs)
        sep = kwargs.get("sep", "")
        return sep.join(ps) if sep else os.path.join(*ps)

    @classmethod
    def get_basename(cls, ext="", prefix="", name="", suffix="", **kwargs):
        """return just a valid file name

        :param ext: string, the extension you want the file to have
        :param prefix: string, this will be the first part of the file's name
        :param name: string, the name you want to use (prefix will be added to the front
            of the name and suffix and ext will be added to the end of the name)
        :param suffix: string, if you want the last bit to be posfixed with something
        :returns: string, the random filename
        """
        basename = name or ""

        if prefix:
            basename = prefix + basename

        if ext:
            if not ext.startswith("."):
                ext = "." + ext

            if basename.endswith(ext):
                basename, ext = os.path.splitext(basename)

        if suffix:
            basename += suffix

        if ext:
            basename += ext

        if not basename:
            raise ValueError("basename is empty")

        return basename

    @classmethod
    def normparts(cls, *parts, **kwargs):
        """Normalize the parts and prepare them to generate the path and value

        :param *parts: list, the parts of the path
        :param **kwargs: anything else needed to generate the parts
        :return: list, a the parts ready to generate path and value
        """
        parts = cls.splitparts(*parts, **kwargs)
        ext = kwargs.pop("ext", "")
        prefix = kwargs.pop("prefix", "")
        suffix = kwargs.pop("suffix", kwargs.pop("postfix", ""))
        name = cls.get_basename(
            ext=ext,
            prefix=prefix,
            name=parts[-1] if parts else "",
            suffix=suffix,
            **kwargs
        )

        if name:
            if parts:
                parts[-1] = name
            else:
                parts.append(name)

        return parts

    @classmethod
    def normpath(cls, *parts, **kwargs):
        """normalize a path, accounting for things like windows dir seps

        the path is the actual file path located in .path

        :param *parts: list, the already normalized parts of the path
        :param **kwargs: anything else
        :return: str, the full path ready to be placed in .path
        """
        path = ""
        if parts:
            path = cls.joinparts(*parts, **kwargs)
            path = os.path.abspath(os.path.expandvars(os.path.expanduser(path)))

        return path

    @classmethod
    def normvalue(cls, *parts, **kwargs):
        """normalize a value 

        the value is the actual string value

        :param *parts: list, the already normalized parts of the path
        :param **kwargs: anything else
        :return: str, the string value
        """
        return kwargs["path"]

    def __new__(cls, *parts, **kwargs):
        parts = cls.normparts(*parts, **kwargs)
        path = cls.normpath(*parts, **kwargs)
        value = cls.normvalue(*parts, path=path, **kwargs)
        path_class = kwargs.pop("path_class", None) # has to be None so create_as works

        instance = super(Path, cls).__new__(
            path_class if path_class else cls,
            value,
            encoding=kwargs.pop("encoding", environ.ENCODING),
            errors=kwargs.pop("errors", environ.ENCODING_ERRORS),
        )
        return cls.create_as(
            instance, 
            path=path,
            parts=parts,
            path_class=path_class,
            **kwargs
        )

    def as_class(self, **kwargs):
        kwargs.setdefault("encoding", self.encoding)
        kwargs.setdefault("errors", self.errors)
        return kwargs["path_class"](self.path, **kwargs)

    def as_file(self, **kwargs):
        """return a new instance of this path as a Filepath"""
        kwargs["path_class"] = self.file_class()
        return self.as_class(**kwargs)
        #return self.create_file(self.path, **kwargs)

    def as_dir(self, **kwargs):
        """return a new instance of this path as a Dirpath"""
        kwargs["path_class"] = self.dir_class()
        return self.as_class(**kwargs)
        #return self.create_dir(self.path, **kwargs)

    def as_path(self):
        """return a new instance of this path as a Path"""
        kwargs["path_class"] = self.path_class()
        return self.as_class(**kwargs)
        #return self.create_path(self.path, **kwargs)

    def as_uri(self):
        """Represent the path as a file URI

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.as_uri
        """
        return "file://{}".format(self.path)

    def empty(self):
        """Return True if the file/directory is empty"""
        raise NotImplementedError()

    def stat(self):
        """Return a os.stat_result object containing information about this path,
        like os.stat(). The result is looked up at each call to this method

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.stat

        :returns: os.stat_result tuple, https://docs.python.org/3/library/os.html#os.stat_result
            0: st_mode
            1: st_ino
            2: st_dev
            3: st_nlink
            4: st_uid
            5: st_gid
            6: st_size
            7: st_atime
            8: st_mtime
            9: st_ctime
        """
        if is_py2:
            ret = os.stat(self.path)

        else:
            ret = self.pathlib.stat()

        return ret

    def chmod(self, mode):
        """Change the file mode and permissions, like os.chmod()

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.chmod
        """
        if isinstance(mode, int):
            mode = "{0:04d}".format(mode)

        try:
            mode = int(mode, 8)
        except TypeError:
            pass

        return os.chmod(self.path, mode)

    def expanduser(self):
        """Return a new path with expanded ~ and ~user constructs, as returned
        by os.path.expanduser()

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.expanduser
        """
        return self

    def owner(self):
        """Return the name of the user owning the file. KeyError is raised if the file’s
        uid isn’t found in the system database.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.owner
        """
        if is_py2:
            import pwd
            return String(pwd.getpwuid(self.stat().st_uid).pw_name)

        else:
            ret = self.pathlib.owner()

        return ret

    def group(self):
        """Return the name of the group owning the file. KeyError is raised if the file’s
        gid isn’t found in the system database.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.group
        """
        if is_py2:
            import grp
            return String(grp.getgrgid(self.stat().st_gid).gr_name)

        else:
            ret = self.pathlib.group()

        return ret

    def exists(self):
        """Whether the path points to an existing file or directory

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.exists
        """
        return os.path.exists(self.path)

    def is_dir(self):
        """Return True if the path points to a directory (or a symbolic link pointing to a directory),
        False if it points to another kind of file.

        False is also returned if the path doesn’t exist or is a broken symlink;
        other errors (such as permission errors) are propagated.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.is_dir
        """
        return os.path.isdir(self.path)

    def isdir(self):
        return self.is_dir()

    def is_file(self):
        """Return True if the path points to a regular file (or a symbolic link pointing to a regular file),
        False if it points to another kind of file.

        False is also returned if the path doesn’t exist or is a broken symlink;
        other errors (such as permission errors) are propagated.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.is_file
        """
        return os.path.isfile(self.path)

    def isfile(self):
        return self.is_file()

    def is_mount(self):
        """Return True if the path is a mount point: a point in a file system where
        a different file system has been mounted. On POSIX, the function checks whether
        path’s parent, path/.., is on a different device than path, or whether path/..
        and path point to the same i-node on the same device — this should detect
        mount points for all Unix and POSIX variants. Not implemented on Windows.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.is_mount
        """
        return os.path.ismount(self.path)

    def is_symlink(self):
        """Return True if the path points to a symbolic link, False otherwise.

        False is also returned if the path doesn’t exist; other errors
        (such as permission errors) are propagated.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.is_symlink
        """
        return os.path.islink(self.path)

    def is_absolute(self):
        """Return whether the path is absolute or not. A path is considered absolute
        if it has both a root and (if the flavour allows) a drive

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.is_absolute
        https://docs.python.org/3/library/os.path.html#os.path.isabs
        """
        return os.path.isabs(self.path)

    def isabs(self):
        return self.is_absolute()

    def is_reserved(self):
        """With PureWindowsPath, return True if the path is considered reserved
        under Windows, False otherwise. With PurePosixPath, False is always returned.

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.is_reserved
        """
        raise NotImplementedError()

    def is_root(self):
        """Return True if path is root (eg, / on Linux), False otherwise"""
        return self.name == ""

    def joinpath(self, *other):
        """Calling this method is equivalent to combining the path with each of the
        other arguments in turn

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.joinpath
        """
        if is_py2:
            ret = self.create(self.path, *other)

        else:
            ret = self.create(self.pathlib.joinpath(*other))

        return ret

    def child(self, *other):
        return self.joinpath(*other)

    def match(self, pattern):
        """Match this path against the provided glob-style pattern. Return True if
        matching is successful, False otherwise.

        If pattern is relative, the path can be either relative or absolute, and matching
        is done from the right

        If pattern is absolute, the path must be absolute, and the whole path must match

        As with other methods, case-sensitivity follows platform defaults

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.match
        """
        if is_py2:
            ret = True
            parts = list(reversed(self.parts))
            patterns = list(reversed(pattern.split("/")))

            for i, part in enumerate(parts):
                if len(patterns) <= i:
                    break

                if part == "/":
                    if patterns[i]:
                        ret = False
                        break

                elif not fnmatch.fnmatch(part, patterns[i]):
                    ret = False
                    break

        else:
            ret = self.pathlib.match(pattern)

        return ret

    def with_name(self, name):
        """Return a new path with the name changed. If the original path doesn’t
        have a name, ValueError is raised

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.with_name
        """
        if self.is_root():
            raise ValueError("{} has an empty name".format(self))
        return self.create(self.parent, name)

    def with_suffix(self, suffix):
        """Return a new path with the suffix changed. If the original path doesn't
        have a suffix, the new suffix is appended instead. If the suffix is an empty string,
        the original suffix is removed

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.with_suffix
        """
        parent = self.parent
        fileroot = self.fileroot
        basename = fileroot

        if not suffix:
            basename = fileroot

        elif not suffix.startswith("."):
            basename += "." + suffix

        else:
            basename += suffix

        return self.create(parent, basename)

    def backup(self, suffix=".bak", ignore_existing=True):
        """backup the file to the same directory with given suffix

        :param suffix: str, what will be appended to the file name (eg, foo.ext becomes
            foo.ext.bak)
        :param ignore_existing: boolean, if True overwrite an existing backup, if false
            then don't backup if a backup file already exists
        :returns: instance, the backup Path
        """
        target = self.create("{}{}".format(self.path, suffix))
        return self.backup_to(target)

    def backup_to(self, target, ignore_existing=True):
        target = self.create(target)
        if ignore_existing or not target.exists():
            self.cp(target)
        return target

    def rename(self, target):
        """Rename this file or directory to the given target, and return a new Path
        instance pointing to target. On Unix, if target exists and is a file,
        it will be replaced silently if the user has permission. target can be
        either a string or another path object

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.rename

        :returns: the new Path instance
        """
        target = self.create_path(target)
        os.rename(self.path, target)
        return self.create(target)

    def replace(self, target):
        """Rename this file or directory to the given target, and return a new Path
        instance pointing to target. If target points to an existing file or directory,
        it will be unconditionally replaced.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.replace

        :returns: the new Path instance.
        """
        if is_py2:
            ret = self.rename(target)

        else:
            target = self.create_path(target)
            self.pathlib.replace(target)
            ret = self.create(target)

        return ret

    def mv(self, target):
        """mimics the behavior of unix mv command"""
        raise NotImplementedError()

    def move_to(self, target):
        return self.mv(target)

    def cp(self, target):
        """mimics the behavior of unix cp command"""
        raise NotImplementedError()

    def copy_to(self, target):
        return self.cp(target)

    def relative_to(self, *other):
        """Compute a version of this path relative to the path represented by other.
        If it’s impossible, ValueError is raised

        returns the relative part to the *other

        :Example:
            d = Path("/foo/bar/baz/che")
            d.relative_to("/foo/bar") # baz/che
            d.relative_to("/foo") # bar/baz/che

        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.relative_to

        :param *other: string|Directory, the directory you want to return that self
            is a child of
        :returns: string, the part of the path that is relative, because these are
            relative paths they can't return Path instances
        """
        if is_py2:
            ancestor_dir = self.joinparts(*other)

            ret = os.path.relpath(self.path, ancestor_dir)
            if ret.startswith(".."):
                raise ValueError("'{}' does not start with '{}'".format(self.path, ancestor_dir))

        else:
            ret = String(self.pathlib.relative_to(*other))

        return ret

    def symlink_to(self, target, target_is_directory=False):
        """Make this path a symbolic link to target. Under Windows, target_is_directory
        must be true (default False) if the link’s target is a directory.
        Under POSIX, target_is_directory’s value is ignored

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.symlink_to
        """
        target = self.create(target)
        if is_py2:
            os.symlink(self.path, target)

        else:
            self.pathlib.symlink_to(target, target_is_directory=target_is_directory)

        return target

    def archive_to(self, target):
        raise NotImplementedError("Waiting for Archivepath api to be finalized")

    def resolve(self, strict=False):
        """Make the path absolute, resolving any symlinks. A new path object is returned

        If the path doesn’t exist and strict is True, FileNotFoundError is raised.
        If strict is False, the path is resolved as far as possible and any remainder
        is appended without checking whether it exists. If an infinite loop is
        encountered along the resolution path, RuntimeError is raised.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.resolve
        """
        ret = self.create(os.path.realpath(self.path))
        if strict:
            if not ret.exists():
                raise OSError(self.path)

        return ret

    def samefile(self, other_path):
        """Return whether this path points to the same file as other_path, which
        can be either a Path object, or a string. The semantics are similar to
        os.path.samefile() and os.path.samestat().

        An OSError can be raised if either file cannot be accessed for some reason.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.samefile
        """
        other_path = self.create(other_path)
        return os.path.samefile(self.path, other_path)

    def touch(self, mode=0o666, exist_ok=True):
        """Create a file at this given path. If mode is given, it is combined with
        the process’ umask value to determine the file mode and access flags.
        If the file already exists, the function succeeds if exist_ok is true
        (and its modification time is updated to the current time), otherwise
        FileExistsError is raised.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.touch
        """
        raise NotImplementedError()

    def rm(self):
        """remove file/dir, does not raise error on file/dir not existing"""
        raise NotImplementedError()

    def delete(self):
        """remove file/dir, alias of .rm(), does not raise error on file/dir not existing"""
        return self.rm()

    def remove(self):
        """remove file/dir, alias of .rm(), does not raise error on file/dir not existing"""
        return self.rm()

    def clear(self):
        """clear the file/directory but don't delete it"""
        raise NotImplementedError()

    def created(self):
        """return a datetime.datetime of when the file was created"""
        t = os.path.getctime(self.path)
        return datetime.datetime.fromtimestamp(t)

    def modified(self):
        """return a datetime.datetime of when the file was modified"""
        # http://stackoverflow.com/a/1526089/5006
        t = os.path.getmtime(self.path)
        return datetime.datetime.fromtimestamp(t)

    def updated(self):
        return self.modified()

    def accessed(self):
        """return a datetime.datetime of when the file was accessed"""
        t = os.path.getatime(self.path)
        return datetime.datetime.fromtimestamp(t)

    def __fspath__(self):
        """conform to PathLike abstract base class for py3.6+

        https://docs.python.org/3/library/os.html#os.PathLike
        """
        return self.path

    def __truediv__(self, other):
        """Synctactic sugar, allows self / "bit" to work"""
        return self.joinpath(other)

    def __div__(self, other): # 2.x
        return self.__truediv__(other)


class Dirpath(Path):
    """Represents a directory so extends Path with methods to iterate through
    a directory"""
    @classmethod
    def cwd(cls):
        """Return a new path object representing the current directory (as returned by os.getcwd())

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.cwd
        """
        return cls.create_dir(os.getcwd())

    @classmethod
    def home(cls):
        """Return a new path object representing the user’s home directory
        (as returned by os.path.expanduser() with ~ construct)

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.home
        """
        return cls.create_dir(os.path.expanduser("~"))

    @classmethod
    def normpaths(cls, paths, baseparts="", **kwargs):
        """normalizes the paths from methods like .add_paths() and .add()

        :param paths: see .add_paths() for description
        :param baseparts: list, will be merged with the paths keys to create the full path
        :returns list of tuples, each tuple will be in the form of (parts, data)
        """
        ret = []
        baseparts = cls.splitparts(baseparts or [], **kwargs)

        if paths:
            if isinstance(paths, Mapping):
                for k, v in paths.items():
                    p = cls.splitparts(baseparts, k, **kwargs)
                    if isinstance(v, Mapping):
                        ret.extend(cls.normpaths(v, p, **kwargs))

                    else:
                        ret.append((p, v))

            elif isinstance(paths, Sequence):
                for k in paths:
                    ret.append((cls.splitparts(baseparts, k, **kwargs), None))

            else:
                raise ValueError("Unrecognized value for paths")

        else:
            ret.append((baseparts, None))

        return ret

    def add_files(self, paths, **kwargs):
        return self.add_paths(paths, **kwargs)

    def add_dirs(self, paths, **kwargs):
        return self.add_paths(paths, **kwargs)

    def add_paths(self, paths, **kwargs):
        """create a whole bunch of files/directories all at once

        :Example:
            Dirpath.add_paths({
                "foo": {
                    "bar": {
                        "baz.txt": "/foo/bar/baz.txt data",
                    },
                    "che": {}, # /foo/che/ directory
                    "bam": None, # /foo/bam/ directory
                }
            })

        :param paths: dict|list
            dict - if paths is a dict, then the keys will be the path part and the
                value will be the data/contents of the file at the full path. If value
                is None or empty dict then that path will be considered a directory.
            list - if paths is a list then it will be a list of directories to create
        :returns: list, all the created Path instances
        """
        ret = []

        ps = self.normpaths(paths, self.path)
        for parts, data in ps:
            if data is None:
                dp = self.create_dir(*parts, **kwargs)
                dp.touch()
                ret.append(dp)

            else:
                fp = self.create_file(parts, **kwargs)
                if data:
                    if isinstance(data, self.path_class()):
                        data.copy_to(fp)

                    elif isinstance(data, (Bytes, bytearray)):
                        fp.write_bytes(data, **kwargs)

                    elif isinstance(data, Str):
                        fp.write_text(data, **kwargs)

                    else:
                        # unknown data is assumed to be something that can be
                        # normalized in .prepare_text()
                        fp.write_text(data, **kwargs)

#                     elif isinstance(data, Sequence):
#                         fp.write_text("\n".join(data), **kwargs)
# 
#                     else:
#                         raise ValueError("Unknown data for {}".format(fp))

                else:
                    fp.touch()

                ret.append(fp)

        return ret

    def add(self, paths, **kwargs):
        """add paths to this directory

        :param paths: dict|list, see @add_paths() for description of paths structure
        :returns: list, all the created Path instances
        """
        return self.add_paths(paths, **kwargs)

    def add_children(self, paths, **kwargs):
        return self.add_paths(paths, **kwargs)

    def add_file(self, target, data="", **kwargs):
        ps = self.add({self.joinparts(target): data}, **kwargs)
        return ps[0]

    def add_dir(self, target):
        ps = self.add({self.joinparts(target): None})
        return ps[0]

    def add_child(self, target, data=None, **kwargs):
        ps = self.add({self.joinparts(target): data}, **kwargs)
        return ps[0]

    def empty(self):
        """Return True if directory is empty"""
        ret = True
        for p in self.iterdir():
            ret = False
            break
        return ret

    def rm(self):
        try:
            shutil.rmtree(self.path)

        except OSError:
            # we don't care if the directory doesn't exist
            pass

    def rmdir(self):
        """Remove this directory. The directory must be empty

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.rmdir
        """
        os.rmdir(self.path)

    def clear(self):
        """Remove all the contents of this directory but leave the directory"""
        for p in self.iterdir():
            p.rm()

    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        """Create a new directory at this given path. If mode is given, it is combined
        with the process’ umask value to determine the file mode and access flags.
        If the path already exists, FileExistsError is raised.

        If parents is true, any missing parents of this path are created as needed;
        they are created with the default permissions without taking mode into account
        (mimicking the POSIX mkdir -p command).

        If parents is false (the default), a missing parent raises FileNotFoundError.

        If exist_ok is false (the default), FileExistsError is raised if the target directory already exists.

        If exist_ok is true, FileExistsError exceptions will be ignored
        (same behavior as the POSIX mkdir -p command), but only if the last path
        component is not an existing non-directory file.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.mkdir
        """
        if not exist_ok:
            if self.is_dir():
                raise OSError("FileExistsError")

        if parents:
            os.makedirs(self.path, mode)

        else:
            os.mkdir(self.path, mode)

    def mv(self, target):
        """Move the entire contents of the directory at self into a directory at target

        :Example:
            $ mv src target
            target/src if target exists
            error is thrown if target/src exists and target/src is not empty
                (mv: rename src to target/src: Directory not empty)
            src is moved to target/src if target/src exists but is empty
            src is moved to target if target does not exist
        """
        target = self.create_dir(target)

        if target.is_dir():
            # https://stackoverflow.com/a/15034373/5006
            target = self.create_dir(target, self.basename)

            if target.is_dir() and not target.empty():
                # this will error out if target is not empty
                raise OSError("Directory not empty: {}".format(target))

        # dir_util says it creates all the parent directories but for some
        # reason it wasn't working, this makes sure the target exist before
        # trying to copy everything over from src (self.path)
        target.touch()
        dir_util.copy_tree(self.path, target, update=1)

        self.rm()

        return target

    def cp(self, target):
        """Copy the entire contents of the directory at self into a directory at target

        :Example:
            $ cp -R src target
            src is copied to target if target does not exist
            target/src if target exists
            src is merged into target/src if target/src exists
        """
        target = self.create_dir(target)

        if target.is_dir():
            target = self.create_dir(target, self.basename)

        target.touch()
        dir_util.copy_tree(self.path, target, update=1)

        return target

    def touch(self, mode=0o666, exist_ok=True):
        """Create the directory at this given path.  If the directory already exists,
        the function succeeds if exist_ok is true (and its modification time is
        updated to the current time), otherwise FileExistsError is raised."""
        if self.exists():
            if not exist_ok:
                raise OSError("FileExistsError")

            os.utime(self.path, None)

        else:
            dir_util.mkpath(self.path)

    def filecount(self, recursive=True):
        """return how many files in directory"""
        return len(list(self.iterfiles(recursive=recursive)))

    def dircount(self, recursive=True):
        """return how many directories in directory"""
        return len(list(self.iterdirs(recursive=recursive)))

    def count(self, recursive=True):
        """return how many files and directories in directory"""
        return self.filecount(recursive=recursive) + self.dircount(recursive=recursive)

    def glob(self, pattern):
        """Glob the given relative pattern in the directory represented by this path,
        yielding all matching files (of any kind)

        The “**” pattern means "this directory and all subdirectories, recursively".
        In other words, it enables recursive globbing

        globs are endswith matches, so if you passed in "*.txt" it would match any
        filepath that ended with ".txt", likewise, if you pass in "bar" it would match
        any files/folders that ended with "bar"

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.glob
        """
        if is_py2:
            # https://docs.python.org/2/library/glob.html
            for fp in glob.iglob(os.path.join(self.path, pattern)):
                yield self.create(fp)

        else:
            for fp in self.pathlib.glob(pattern):
                yield self.create(fp)

    def rglob(self, pattern):
        """recursive glob

        This is like calling Path.glob() with “**/” added in front of the given
        relative pattern, meaning it will match files in this directory and all
        subdirectories

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.rglob
        """
        if is_py2:
            for fp in self.glob(pattern):
                yield fp

            for dp in self.iterdirs(recursive=True):
                for fp in glob.iglob(os.path.join(dp, pattern)):
                    yield self.create(fp)

        else:
            for fp in self.pathlib.rglob(pattern):
                yield self.create(fp)

    def reglob(self, pattern, recursive=True):
        """A glob but uses a regex instead of the common glob syntax

        :param pattern: regex, the pattern to check for
        :param recursive: boolean, True if you would like to check this directory
            and all subdirectories, False if just self directory
        :returns: generator, yields all found Path instances that match
        """
        it = self.rglob("*") if recursive else self.glob("*")
        for p in it:
            if re.search(pattern, p):
                yield p

    def cbglob(self, callback, recursive=True):
        """A glob but uses a callback instead of the common glob syntax

        :param callback: callable, a callback with signature callback(path) that
            returns a boolean
        :param recursive: boolean, True if you would like to check this directory
            and all subdirectories, False if just self directory
        :returns: generator, yields all found Path instances that match
        """
        it = self.rglob("*") if recursive else self.glob("*")
        for p in it:
            if callback(p):
                yield p

    def scandir(self):
        """
        https://docs.python.org/3.5/library/os.html#os.scandir

        https://github.com/benhoyt/scandir
        https://bugs.python.org/issue11406
        """
        # TODO -- make this work similarly for py2
        for entry in os.scandir(self.path):
            yield entry

    def iterdir(self):
        """When the path points to a directory, yield path objects of the directory contents

        This will only yield files/folders found in this directory, it is not recursive

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.iterdir

        :returns: generator, yields children Filepath and Dirpath instances found
            only in this directory
        """
        if is_py2:
            for basename in os.listdir(self.path):
                yield self.create(self.path, basename)

        else:
            for p in self.pathlib.iterdir():
                yield self.create(p)

    def iterdirs(self, recursive=True):
        """iterate through all the directories similar to .iterdir() but only
        returns directories

        :param recursive: boolean, if True then iterate through directories in
            self and all subdirectories
        """
        for basedir, directories, files in os.walk(self.path, topdown=True):
            basedir = self.create_dir(basedir)
            for basename in directories:
                yield self.create_dir(basedir, basename)

            if not recursive:
                break

    def iterfiles(self, recursive=True):
        """iterate through all the files in this directory only"""
        """iterate through all the files in this directory and subdirectories"""
        for basedir, directories, files in os.walk(self.path, topdown=True):
            basedir = self.create_dir(basedir)
            for basename in files:
                yield self.create_file(basedir, basename)

            if not recursive:
                break

    def children(self, pattern="", recursive=True):
        """Syntactic sugar around the other directory iteration methods

        :param pattern: callable|string
        :param recursive: boolean, if True go through all files/folders, if false
            then just go through self directory
        :returns: generator, yields all matching Path instances, or every Path
            instance if pattern is empty
        """
        if pattern:
            if callable(pattern):
                for p in self.cbglob(pattern, recursive=recursive):
                    yield p

            else:
                if recursive:
                    for p in self.rglob(pattern):
                        yield p

                else:
                    for p in self.glob(pattern):
                        yield p

        else:
            it = self.rglob if recursive else self.glob
            for p in it("*"):
                yield p

    def walk(self, *args, **kwargs):
        """passthrough for os.walk

        https://docs.python.org/3/library/os.html#os.walk
        :returns: yields (basedir, dirs, files) in the exact way os.walk does, these
            are not Path instances
        """
        for basedir, dirs, files in os.walk(self.path, *args, **kwargs):
            yield basedir, dirs, files

    def __iter__(self):
        """Like os.walk but will return full Path instances in each tuple

        :returns: tuple, (basedir, dirs, files) where the dirs and files lists have
            full Dirpath and Filepath instances
        """
        for basedir, dirs, files in self.walk(topdown=True):
            basedir = self.create_dir(basedir)
            for i in range(len(dirs)):
                dirs[i] = self.create_dir(basedir, dirs[i])

            for i in range(len(files)):
                files[i] = self.create_file(basedir, files[i])

            yield basedir, dirs, files

    def has(self, pattern="", recursive=True):
        """Check for pattern in directory

        :Example:
            d = Dirpath("foo")
            d.add_file("bar/che.txt", "che.txt data")

            d.has("che.txt") # True
            d.has("bar") # True
            d.has("bar/che.txt") # True
            d.has("bar/che") # False
            d.has("bar/che.*") # True

        :param pattern: string, the subpath or pattern. If pattern is empty then
            it will return True if directory has any file/folder
        :returns: boolean, True if the pattern is in the directory
        """
        for p in self.children(pattern, recursive=recursive):
            return True
        return False


class Filepath(Path):
    """Represents a file so extends Path with file reading/writing methods"""
    def empty(self):
        """Return True if file is empty"""
        ret = True
        if self.exists():
            st = self.stat()
            ret = False if st.st_size else True
        return ret

    def open(self, mode='r', buffering=-1, encoding=None, errors=None, newline=None):
        """Open the file pointed to by the path, like the built-in open() function does

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.open
        """
        if not mode:
            mode = "r" if encoding else "rb"

        try:
            if is_py2:
                if encoding:
                    fp = codecs.open(
                        self.path,
                        mode=mode,
                        encoding=encoding,
                        errors=errors,
                        buffering=buffering,
                    )

                else:
                    fp = open(
                        self.path,
                        mode=mode,
                        buffering=buffering,
                    )

            else:
                fp = open(
                    self.path,
                    mode=mode,
                    encoding=encoding,
                    errors=errors,
                    buffering=buffering,
                    newline=newline,
                )

        except IOError:
            if self.exists():
                raise

            else:
                self.touch()
                fp = self.open(mode, buffering, encoding, errors, newline)

        return fp

    def open_text(self, mode='r', encoding=None, errors=None, **kwargs):
        """Just like .open but will set encoding and errors to class values
        if they aren't passed in"""
        encoding = encoding or self.encoding
        errors = errors or self.errors
        return self.open(mode, encoding=encoding, errors=errors, **kwargs)

    def __call__(self, mode="w+", **kwargs):
        """Allow an easier interface for opening a writing file descriptor

        This uses the class defaults for things like encoding, so it's better to
        use .open() or .write_bytes() if you want to write non-encoded raw text

        :Example:
            p = Filepath("foo/bar.ext")
            with p("a+) as fp:
                fp.write("some value")
        """
        return self.open(
            mode=mode,
            buffering=kwargs.get("buffering", -1),
            newline=kwargs.get("newline", None),
            encoding=kwargs.get("encoding", self.encoding),
            errors=kwargs.get("errors", self.errors),
        )

    def __enter__(self):
        """Allow an easier interface for opening a writing file descriptor

        :Example:
            p = Filepath("foo/bar.ext")
            with p as fp:
                fp.write("some value")
        """
        self._context_fp = self()
        return self._context_fp

    def __exit__(self, exception_type, exception_val, trace):
        self._context_fp.close()
        del self._context_fp

    def read_bytes(self):
        """Return the binary contents of the pointed-to file as a bytes object

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.read_bytes
        """
        # https://stackoverflow.com/a/32957860/5006
        with self.open(mode="rb") as fp:
            return fp.read()

    def read_text(self, encoding=None, errors=None):
        """Return the decoded contents of the pointed-to file as a string

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.read_text
        """
        encoding = encoding or self.encoding
        errors = errors or self.errors
        with self.open(encoding=encoding, errors=errors) as fp:
            return fp.read()

    def prepare_bytes(self, data, **kwargs):
        """Internal method used to prepare the data to be written

        :param data: str, the text that will be written
        :param **kwargs: keywords, you can pass in encoding here
        :returns: tuple, (data, encoding, errors)
        """
        encoding = kwargs.get("encoding", None) or getattr(data, "encoding", self.encoding)
        errors = kwargs.get("errors", None) or getattr(data, "errors", self.errors)
        data = ByteString(data, encoding=encoding, errors=errors)
        return data, encoding, errors

    def prepare_text(self, data, **kwargs):
        """Internal method used to prepare the data to be written

        :param data: str, the text that will be written
        :param **kwargs: keywords, you can pass in encoding here
        :returns: tuple, (data, encoding, errors)
        """
        encoding = kwargs.get("encoding", None) or getattr(data, "encoding", self.encoding)
        errors = kwargs.get("errors", None) or getattr(data, "errors", self.errors)
        data = String(data, encoding=encoding, errors=errors)
        return data, encoding, errors

    def write_bytes(self, data, **kwargs):
        """Open the file pointed to in bytes mode, write data to it, and close the file

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.write_bytes

        NOTE -- having **kwargs means the interface is different than Pathlib.write_bytes

        :param data: bytes
        :param **kwargs: supports errors and encoding keywords to convert data to
            bytes
        """
        data, encoding, errors = self.prepare_bytes(data, **kwargs)
        with self.open(mode="wb+") as fp:
            return fp.write(data)

    def write_text(self, data, **kwargs):
        """Open the file pointed to in text mode, write data to it, and close the file

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.write_text

        :param data: str
        :param **kwargs: supports errors and encoding keywords to convert data to
            str/unicode
        """
        data, encoding, errors = self.prepare_text(data, **kwargs)
        with self.open(mode="w+", encoding=encoding, errors=errors) as fp:
            return fp.write(data)

    def write(self, data):
        """Write data as either bytes or text

        :param data: bytes|string
        :return: the amount written
        """
        if isinstance(data, (Bytes, bytearray)):
            return self.write_bytes(data)

        else:
            return self.write_text(data)

    def append_text(self, data):
        data, encoding, errors = self.prepare_text(data)
        with self.open(mode="a+", encoding=encoding, errors=errors) as fp:
            return fp.write(data)

    def joinpath(self, *other):
        raise NotImplementedError()

    def cp(self, target):
        """copy self to/into target"""
        target = self.create(target)
        if target.is_dir():
            target = self.create_file(target, self.basename)

        shutil.copy(self.path, target)
        return target.as_file()

    def mv(self, target):
        target = self.create(target)
        if target.is_dir():
            target = self.create_file(target, self.basename)

        shutil.move(self.path, target)
        return target.as_file()

    def unlink(self, missing_ok=False):
        """Remove this file or symbolic link. If the path points to a directory,
        use Path.rmdir() instead.

        If missing_ok is false (the default), FileNotFoundError is raised if the
        path does not exist.

        If missing_ok is true, FileNotFoundError exceptions will be ignored
        (same behavior as the POSIX rm -f command).

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.unlink
        """
        try:
            os.unlink(self.path)

        except OSError as e:
            if not missing_ok:
                raise

    def rm(self):
        return self.unlink(missing_ok=True)

    def clear(self):
        """clear the file/directory but don't delete it"""
        self.write_bytes(b"")

    def touch(self, mode=0o666, exist_ok=True):
        """Create a file at this given path. If mode is given, it is combined with
        the process’ umask value to determine the file mode and access flags.
        If the file already exists, the function succeeds if exist_ok is true
        (and its modification time is updated to the current time), otherwise
        FileExistsError is raised.

        https://docs.python.org/3/library/pathlib.html#pathlib.Path.touch
        """
        if self.exists():
            if not exist_ok:
                raise OSError("FileExistsError")

            os.utime(self.path, None)

        else:
            self.parent.touch(mode, exist_ok)

            # http://stackoverflow.com/a/1160227/5006
            with self.open("a") as f:
                os.utime(self.path, None)

    def linecount(self):
        """return line count"""
        return len(list(self.splitlines()))

    def lc(self):
        return self.linecount()

    def count(self):
        """how many characters in the file"""
        return len(self.read_text())

    def checksum(self):
        """return md5 hash of a file"""
        h = hashlib.md5()
        blocksize = 65536
        # http://stackoverflow.com/a/21565932/5006
        with self.open(mode="rb") as fp:
            for block in iter(lambda: fp.read(blocksize), b""):
                h.update(block)
        return h.hexdigest()

    def hash(self): return self.checksum()
    def md5(self): return self.checksum()

    def head(self, count):
        """get the first count lines of self.path

        :param count: int, how many lines you want from the start of the file
        :returns: list, the lines in a similar format to .lines()
        """
        if count == 0:
            ret = self.splitlines()
        else:
            ret = [l[1] for l in enumerate(self.splitlines()) if l[0] < count]
        return ret

    def tail(self, count):
        """
        get the last count lines of self.path

        https://stackoverflow.com/a/280083/5006

        :param count: int, how many lines you want from the end of the file
        :returns: list, the lines in a similar format to .lines()
        """
        if count == 0:
            ret = self.splitlines()
        else:
            ret = deque(self.splitlines(), maxlen=count)
        return ret

    def splitlines(self, keepends=False):
        """iterate through all the lines"""
        with self.open("r", encoding=self.encoding, errors=self.errors) as f:
            for line in f:
                yield line if keepends else line.rstrip()

    def __iter__(self):
        for line in self.splitlines():
            yield line

    def has(self, pattern=""):
        """Check for pattern in the body of the file

        :Example:
            d = Filepath("foo.txt")

            d.has("<TEXT>") # True

        :param pattern: string|callable, the contents in the file. If callable then
            it will do pattern(line) for each line in the file
        :returns: boolean, True if the pattern is in the file
        """
        with self.open("r", encoding=self.encoding, errors=self.errors) as f:
            if pattern:
                if callable(pattern):
                    for line in f:
                        if pattern(line):
                            return True

                else:
                    return pattern in f.read()

            else:
                data = f.read(1)
                return True if data else False

        return False


class Archivepath(Dirpath):
    """This was based off of code from herd.path but as I was expanding it I realized
    it was going to be more work than my needs currently warranted so I'm leaving
    this here and I'll get back to it another time


    https://docs.python.org/3/library/tarfile.html 
    https://docs.python.org/2/library/tarfile.html#tarfile-objects

    https://docs.python.org/3/library/zipfile.html
    https://docs.python.org/2/library/zipfile.html#zipfile-objects
    """
    def __new__(cls, *parts, **kwargs):
        instance = super(Archivepath, cls).__new__(cls, *parts, **kwargs)

        archive_info = {
            ".zip": {
                "archive_class": zipfile.ZipFile,
                "archive_format": "zip",
                "write_mode": "w",
                "write_method": "write",
                "__iter__": "namelist",
            },
            ".tar": {
                "archive_class": tarfile.TarFile,
                "archive_format": "tar",
                "write_mode": "w:",
                "write_method": "add"
            },
            ".tar.gz": {
                "archive_class": tarfile.TarFile,
                "archive_format": "gztar",
                "write_mode": "w:gz",
                "write_method": "add"
            },
            ".tar.bz2": {
                "archive_class": tarfile.TarFile,
                "archive_format": "bztar",
                "write_mode": "w:bz2",
                "write_method": "add"
            },
            ".tar.xz": {
                "archive_class": tarfile.TarFile,
                "archive_format": "xztar",
                "write_mode": "w:xz",
                "write_method": "add"
            },
        }

        instance.info = {}
        for suffix, info in archive_info.items():
            if instance.endswith(suffix):
                instance.info = info
                break

        if not instance.info:
            raise ValueError("No archive info found for archive file {}".format(instance.path))

        return instance

    def __iter__(self):
        z = self.info["archive_class"](self.path)
        for n in getattr(z, self.info["__iter__"])():
            yield n

    def add(self, target, arcname=""):
        target = self.create(target)

        # !!! This works but it overwrites the zip file on every call, a
        # solution would be to make this a context manager so you could do
        # something like:
        #
        # with self as z:
        #     z.add(path1)
        #     z.add(path2)
        #     z.add(path3)
        #
        # so each .add copies the files/directories to a temp directory and when the __exit__
        # fires you would use shutil.make_archive to create the actual archive
        # file, this method would then be renamed .set()


        # if the archive doesn't exist we want to make sure at least the directory
        # exists because the file will be created when target is added to the
        # archive
        if not self.exists():
            self.parent.touch()

        if target.is_file():
            if not arcname:
                arcname = target.basename

            with self.info["archive_class"](self.path, mode=self.info["write_mode"]) as z:
                getattr(z, self.info["write_method"])(target, arcname=arcname)

        elif target.is_dir():
            if not arcname:
                arcname = target.path

            target = self.stempath
            # https://docs.python.org/3/library/shutil.html#shutil.make_archive
            # https://stackoverflow.com/a/25650295/5006
            shutil.make_archive(self.stempath, self.info["archive_format"], root_dir=arcname)

        else:
            raise ValueError("Target is neither a file or a directory")

    def extract_to(self, target):
        target = self.create_dir(target)


class TempPath(object):
    basedir = None

    @property
    def relpath(self):
        return self.relative_to(self.basedir)

    @property
    def relparts(self):
        parts = self.splitparts(self.relpath)
        return parts

    @classmethod
    def gettempdir(cls):
        """return the system temp directory"""
        return tempfile.gettempdir()

    @classmethod
    def mktempdir(cls, **kwargs):
        """pass through for tempfile.mkdtemp, this is here so it can be overridden
        in child classes and customized

        :param **kwargs:
            - prefix
            - suffix
            - dir
        :returns: string, the directory path
        """
        return tempfile.mkdtemp(**kwargs)

    @classmethod
    def get_basename(cls, ext="", prefix="", name="", suffix="", **kwargs):
        """return just a valid file name

        :param ext: string, the extension you want the file to have
        :param prefix: string, this will be the first part of the file's name
        :param suffix: string, if you want the last bit to be posfixed with something
        :param name: string, the name you want to use (prefix will be added to the front
            of the name and ext will be added to the end of the name)
        :returns: string, the random filename
        """
        if not name or name == "/":
            name = "".join(random.sample(string.ascii_letters, random.randint(3, 11))).lower()

        return super(TempPath, cls).get_basename(
            ext=ext,
            prefix=prefix,
            name=name,
            suffix=suffix,
            **kwargs
        )

    @classmethod
    def get_parts(cls, count=1, prefix="", name="", suffix="", ext="", **kwargs):
        """Returns count parts

        :param count: int, how many parts you want in your module path (1 is foo, 2 is foo, bar, etc)
        :param prefix: string, if you want the last bit to be prefixed with something
        :param suffix: string, if you want the last bit to be posfixed with something
        :param name: string, the name you want to use for the last bit
            (prefix will be added to the front of the name and postfix will be added to
            the end of the name)
        :returns: list
        """
        parts = []
        count = max(count, 1)

        for x in range(count - 1):
            parts.append(cls.get_basename())

        parts.append(cls.get_basename(prefix=prefix, name=name, suffix=suffix, ext=ext, **kwargs))
        return parts

    @classmethod
    def tempdir_class(cls):
        return TempDirpath

    @classmethod
    def tempfile_class(cls):
        return TempFilepath

    @classmethod
    def create_tempfile(cls, *parts, **kwargs):
        kwargs["path_class"] = cls.tempfile_class()
        return cls.create(*parts, **kwargs)

    @classmethod
    def create_tempdir(cls, *parts, **kwargs):
        kwargs["path_class"] = cls.tempdir_class()
        return cls.create(*parts, **kwargs)


# !!! Ripped from herd.path
class TempDirpath(TempPath, Dirpath):
    """Create a temporary directory

    https://docs.python.org/3/library/tempfile.html

    :Example:
        d = TempDirpath("foo", "bar")
        print(d) # $TMPDIR/foo/bar
    """

#     @classmethod
#     def add_paths(cls, paths, baseparts="", **kwargs):
#         # this check is to normalize the base path to use a temporary dirpath if
#         # it doesn't already have one
#         if baseparts:
#             baseparts = [cls.create_tempdir(), baseparts]
# 
#         else:
#             if isinstance(paths, Mapping):
#                 full_path = False
#                 tp = cls.gettempdir()
#                 for k in paths.keys():
#                     if k.startswith(tp):
#                         full_path = True
#                         break
# 
#                 if not full_path:
#                     baseparts = cls.create_tempdir()
# 
#         return super(TempDirpath, cls).add_paths(paths, baseparts, **kwargs)


    @classmethod
    def normparts(cls, *parts, **kwargs):
        # https://docs.python.org/3/library/tempfile.html#tempfile.mkdtemp
        suffix = kwargs.pop("suffix", kwargs.pop("postfix", ""))
        prefix = kwargs.pop("prefix", "")
        basedir = kwargs.pop("dir", "")

        parts = list(filter(None, parts))
        if parts:
            parts = super(TempDirpath, cls).normparts(*parts, **kwargs)

        if basedir:
            parts = [basedir] + list(parts)

        else:
            create_dir = True
            if parts:
                path = cls.joinparts(*parts)
                if os.path.isabs(path) and os.path.isdir(path):
                    create_dir = False

            if create_dir:
                basedir = cls.mktempdir(
                    suffix=suffix,
                    prefix=prefix,
                )
                parts = [basedir] + list(parts)


        return parts





#     @classmethod
#     def normparts(cls, *parts, **kwargs):
#         # https://docs.python.org/3/library/tempfile.html#tempfile.mkdtemp
#         suffix = kwargs.pop("suffix", kwargs.pop("postfix", ""))
#         prefix = kwargs.pop("prefix", "")
#         basedir = kwargs.pop("dir", "")
#         if not basedir:
#             basedir = tempfile.mkdtemp(
#                 suffix=suffix,
#                 prefix=prefix,
#                 #dir=tmpdir, # cls.gettempdir()
#             )
# 
#         parts = list(filter(None, parts))
#         if parts:
#             parts = super(TempDirpath, cls).normparts(*parts, **kwargs)
# 
#         return [basedir] + list(parts)

    @classmethod
    def create_as(cls, instance, **kwargs):
        kwargs.setdefault("touch", True)
        instance.basedir = kwargs["parts"][0]
        instance = super(TempDirpath, cls).create_as(instance, **kwargs)
        return instance
Dirtemp = TempDirpath
Tempdir = TempDirpath


class TempFilepath(TempPath, Filepath):
    """Create a temporary file

    :Example:
        f = TempFilepath("foo", "bar.ext")
        print(f) # $TMPDIR/foo/bar.ext
    """
#     @classmethod
#     def normparts(cls, *parts, **kwargs):
#         parts = super(TempFilepath, cls).normparts(*parts, **kwargs)
#         basedir = TempDirpath(dir=kwargs.pop("dir", ""))
#         return [basedir] + parts


    @classmethod
    def normparts(cls, *parts, **kwargs):
        parts = super(TempFilepath, cls).normparts(*parts, **kwargs)

        basedir = kwargs.pop("dir", "")
        if basedir:
            parts = [cls.create_tempdir(dir=basedir)] + parts

        else:
            # check to see if parts is a complete path
            path = cls.joinparts(*parts)
            if not os.path.isabs(path) or not os.path.isfile(path):
                parts = [cls.create_tempdir()] + parts

        return parts



#     @classmethod
#     def normparts(cls, *parts, **kwargs):
#         parts = super(TempFilepath, cls).normparts(*parts, **kwargs)
#         basedir = kwargs.pop("dir", "")
#         if basedir:
#             basedir = TempDirpath(dir=basedir)
#             parts = [basedir] + parts
#         return parts
#         #return [basedir] + parts

    @classmethod
    def create_as(cls, instance, **kwargs):
        kwargs.setdefault("touch", True)
        instance.basedir = kwargs["parts"][0]
        instance = super(TempFilepath, cls).create_as(instance, **kwargs)

        data = kwargs.pop("data", kwargs.pop("contents", None))
        if data:
            instance.write(data)

        return instance
Filetemp = TempFilepath
Tempfile = TempFilepath


class Cachepath(Filepath):
    """A file that can contain cached data

    :Example:
        c = Cachepath("key")
        if c:
            data = c.read()
        else:
            data = do_something_long()
            c.write(data)
    """
    prefix = ''
    """set the key prefix"""

    ttl = 0
    """how long to cache the result in seconds, 0 for unlimited"""

    def __new__(cls, *keys, **kwargs):
        if not keys:
            raise ValueError("*keys was empty")

        ttl = kwargs.pop("ttl", cls.ttl)
        prefix = kwargs.pop("prefix", cls.prefix)

        basedir = kwargs.get("dir", "")
        if not basedir:
            basedir = environ.CACHE_DIR

        key = cls.create_key(*keys, prefix=prefix)
        instance = super(Cachepath, cls).__new__(cls, basedir, key)

        instance.ttl = ttl
        instance.prefix = prefix
        return instance

    @classmethod
    def create_key(cls, *keys, **kwargs):
        parts = []
        prefix = kwargs.pop("prefix", cls.prefix)
        if prefix:
            parts.append(prefix)
        parts.extend(keys)

        badchars = String(string.punctuation).stripall(".-_")

        keys = map(lambda s: String(s).stripall(badchars), parts)
        return '.'.join(keys)

    def __nonzero__(self):
        ret = True
        if self.empty():
            ret = False

        else:
            if self.ttl:
                ret = self.modified_within(seconds=self.ttl)

        return ret

    def __bool__(self):
        return self.__nonzero__()

    def write(self, data):
        b = pickle.dumps(data, pickle.HIGHEST_PROTOCOL)
        self.write_bytes(b)

    def read(self):
        return pickle.loads(self.read_bytes())

    def modified_within(self, seconds=0, **timedelta_kwargs):
        """returns true if the file has been modified within the last seconds

        :param seconds: integer, how many seconds
        :param **timedelta_kwargs: dict, can be thing like days, must be passed
            as a named keyword
        :returns: boolean, True if file has been modified after the seconds back
        """
        now = datetime.datetime.now()
        then = self.modified()
        timedelta_kwargs["seconds"] = seconds
        td_check = datetime.timedelta(**timedelta_kwargs)
        #pout.v(now, td_check, then)
        return (now - td_check) < then


class Sentinel(Cachepath):
    """Creates a file after the first failed boolean check, handy when you only want
    to check things at certain intervals

    :Example:
        s = Sentinel("foo")
        if not s:
            # do something you don't want to do all the time
        if not s:
            # this won't ever be ran because the last `if not s` check will have
            # created the sentinel file
    """
    def __new__(cls, *keys, **kwargs):
        """Create the sentinel file

        :param *keys: same arguments as Cachepath
        :param **kwargs:
            yearly: bool, create the file once per year
            monthly: bool, create the file once per month
            weekly: bool, create the file once per week
            daily: bool, create the file once per day
            hourly: bool, create the file once per hour
            ttl: bool, create the file every ttl seconds
        """
        now = datetime.datetime.utcnow()
        year = now.strftime("%Y")
        month = now.strftime("%m")
        day = now.strftime("%d")
        week = now.strftime("%W")
        hour = now.strftime("%H")

        dates = []
        if kwargs.pop("yearly", False):
            dates = [year]

        elif kwargs.pop("monthly", False):
            dates = [year, month]

        elif kwargs.pop("weekly", False):
            dates = [year, week]

        elif kwargs.pop("daily", False):
            dates = [year, month, day]

        elif kwargs.pop("hourly", False):
            dates = [year, month, day, hour]

        keys = list(keys) + dates

        instance = super(Sentinel, cls).__new__(cls, *keys, **kwargs)
        return instance

    def __nonzero__(self):
        ret = False
        if self.exists():
            ret = True
            if self.ttl:
                ret = not self.modified_within(seconds=self.ttl)

        else:
            # we create the file after the first failed exists check
            self.touch()

        return ret


# !!! Ripped from pyt.path which was ripped from pout.path
class SitePackagesDirpath(Dirpath):
    """Finds the site-packages directory and sets the value of this string to that
    path"""
    def __new__(cls):
        basepath = cls._basepath
        if not basepath:
            try:
                paths = site.getsitepackages()
                basepath = paths[0] 
                logger.debug(
                    "Found site-packages directory {} using site.getsitepackages".format(
                        basepath
                    )
                )

            except AttributeError:
                # we are probably running this in a virtualenv, so let's try a different
                # approach
                # try and brute-force discover it since it's not defined where it
                # should be defined
                sitepath = os.path.join(os.path.dirname(site.__file__), "site-packages")
                if os.path.isdir(sitepath):
                    basepath = sitepath
                    logger.debug(
                        "Found site-packages directory {} using site.__file__".format(
                            basepath
                        )
                    )

                else:
                    for path in sys.path:
                        if path.endswith("site-packages"):
                            basepath = path
                            logger.debug(
                                "Found site-packages directory {} using sys.path".format(
                                    basepath
                                )
                            )
                            break

                    if not basepath:
                        for path in sys.path:
                            if path.endswith("dist-packages"):
                                basepath = path
                                logger.debug(
                                    "Found dist-packages directory {} using sys.path".format(
                                        basepath
                                    )
                                )
                                break

        if not basepath:
            raise IOError("Could not find site-packages directory")

        return super(SitePackagesDirpath, cls).__new__(cls, basepath)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
from wsgiref.headers import Headers as BaseHeaders
import itertools

from .compat import *
from .copy import Deepcopy
from .string import String, ByteString


class Headers(BaseHeaders, Mapping):
    """handles headers, see wsgiref.Headers link for method and use information

    Handles normalizing of header names, the problem with headers is they can
    be in many different forms and cases and stuff (eg, CONTENT_TYPE and Content-Type),
    so this handles normalizing the header names so you can request Content-Type
    or CONTENT_TYPE and get the same value.

    This has the same interface as Python's built-in wsgiref.Headers class but
    makes it even more dict-like and will return titled header names when iterated
    or anything (eg, Content-Type instead of all lowercase content-type)

    http headers spec:
        https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html


    wsgiref class docs:
        https://docs.python.org/2/library/wsgiref.html#module-wsgiref.headers
        https://hg.python.org/cpython/file/2.7/Lib/wsgiref/headers.py
    actual python3 code:
        https://github.com/python/cpython/blob/master/Lib/wsgiref/headers.py
    """
    encoding = "iso-8859-1"
    """From rfc2616:

        The default language is English and the default character set is ISO-8859-1.

        If a character set other than ISO-8859-1 is used, it MUST be encoded in the
        warn-text using the method described in RFC 2047"""

    def __init__(self, headers=None, **kwargs):
        super(Headers, self).__init__([])
        self.update(headers, **kwargs)

    def _convert_string_part(self, bit):
        """each part of a header will go through this method, this allows further
        normalization of each part, so a header like FOO_BAR would call this method
        twice, once with foo and again with bar

        :param bit: string, a part of a header all lowercase
        :returns: string, the normalized bit
        """
        if bit == "websocket":
            bit = "WebSocket"
        else:
            bit = bit.title()
        return bit

    def _convert_string_name(self, k):
        """converts things like FOO_BAR to Foo-Bar which is the normal form"""
        k = String(k, self.encoding)
        bits = k.lower().replace('_', '-').split('-')
        return "-".join((self._convert_string_part(bit) for bit in bits))

    def _convert_string_type(self, v):
        """Override the internal method wsgiref.headers.Headers uses to check values
        to make sure they are strings"""
        # wsgiref.headers.Headers expects a str() (py3) or unicode (py2), it
        # does not accept even a child of str, so we need to convert the String
        # instance to the actual str, as does the python wsgi methods, so even
        # though we override this method we still return raw() strings so we get
        # passed all the type(v) == "str" checks
        # sadly, this method is missing in 2.7
        # https://github.com/python/cpython/blob/2.7/Lib/wsgiref/headers.py
        return String(v).raw()

    def get_all(self, name):
        name = self._convert_string_name(name)
        return super(Headers, self).get_all(name)

    def get(self, name, default=None):
        name = self._convert_string_name(name)
        return super(Headers, self).get(name, default)

    def __delitem__(self, name):
        name = self._convert_string_name(name)
        return super(Headers, self).__delitem__(name)

    def __setitem__(self, name, val):
        name = self._convert_string_name(name)
        if is_py2:
            val = self._convert_string_type(val)
        return super(Headers, self).__setitem__(name, val)

    def setdefault(self, name, val):
        name = self._convert_string_name(name)
        if is_py2:
            val = self._convert_string_type(val)
        return super(Headers, self).setdefault(name, val)

    def add_header(self, name, val, **params):
        name = self._convert_string_name(name)
        if is_py2:
            val = self._convert_string_type(val)
        return super(Headers, self).add_header(name, val, **params)

    def keys(self):
        return [k for k, v in self._headers]

    def items(self):
        for k, v in self._headers:
            yield k, v

    def iteritems(self):
        return self.items()

    def iterkeys(self):
        for k in self.keys():
            yield k

    def __iter__(self):
        for k, v in self._headers:
            yield k

    def pop(self, name, *args, **kwargs):
        """remove and return the value at name if it is in the dict

        This uses *args and **kwargs instead of default because this will raise
        a KeyError if default is not supplied, and if it had a definition like
        (name, default=None) you wouldn't be able to know if default was provided
        or not

        :param name: string, the key we're looking for
        :param default: mixed, the value that would be returned if name is not in
            dict
        :returns: the value at name if it's there
        """
        val = self.get(name)
        if val is None:
            if args:
                val = args[0]
            elif "default" in kwargs:
                val = kwargs["default"]
            else:
                raise KeyError(name)

        else:
            del self[name]

        return val

    def update(self, headers, **kwargs):
        if not headers: headers = {}
        if isinstance(headers, Mapping):
            headers.update(kwargs)
            headers = headers.items()

        else:
            if kwargs:
                headers = itertools.chain(
                    headers,
                    kwargs.items()
                )

        for k, v in headers:
            self[k] = v

    def copy(self):
        return Deepcopy().copy(self)

    def list(self):
        """Return all the headers as a list of headers instead of a dict"""
        return [": ".join(h) for h in self.items() if h[1]]


class Environ(Headers):
    """just like Headers but allows any values (headers converts everything to unicode
    string)"""
    def _convert_string_type(self, v):
        return v





# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import logging

from .string import (
    String,
    ByteString,
    Base64,
    HTMLCleaner,
)
from .number import (
    Shorten,
    Integer,
    Hex,
    Binary,
)
from .collections import (
    Pool,
    PriorityQueue,
    NormalizeDict,
    idict,
    AppendList,
    OrderedList,
    Trie,
)
from .csv import (
    CSV,
)
from .enum import (
    Enum,
)
from .path import (
    Path,
    Filepath,
    Dirpath,
    TempFilepath, Filetemp,
    TempDirpath, Dirtemp,
    Cachepath,
    Sentinel,
    SitePackagesDirpath,
)
from .environ import (
    Environ,
)
from .url import (
    Url,
    Host,
)
from .copy import (
    Deepcopy,
)
from .datetime import (
    Datetime,
)
from .headers import (
    Headers,
    Environ,
)
from .token import (
    # we don't import Token and SubToken because their names are too generic for
    # toplevel, if you need them import them directly from the submodule
    StreamTokenizer,
    Tokenizer,
)


__version__ = "0.1.0"


# get rid of "No handler found" warnings (cribbed from requests)
logging.getLogger(__name__).addHandler(logging.NullHandler())




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import datetime
import inspect
import logging
import re

from .compat import *
from .string import String


logger = logging.getLogger(__name__)


class Datetime(datetime.datetime):
    """Wrapper around standard datetime.datetime class that assures UTC time and
    full ISO8601 date strings with Z timezone.

    You can create an object multiple ways:

        d = Datetime() # current utc time
        d = Datetime("2019-10-01T23:26:22.079296Z") # parses ISO8601 date from string
        d = Datetime(2019, 10, 1) # standard creation syntax

    https://docs.python.org/3/library/datetime.html#datetime-objects
    """
    FORMAT_ISO_8601 = "%Y-%m-%dT%H:%M:%S.%fZ"
    FORMAT_PRECISION_SECONDS = "%Y-%m-%dT%H:%M:%SZ"
    FORMAT_PRECISION_DAY = "%Y-%m-%d"

    @property
    def yearname(self):
        """return the full 4 digit year"""
        return self.strftime("%Y")

    @property
    def monthname(self):
        """return the full month name"""
        return self.strftime("%B")

    @property
    def dayname(self):
        """return the full day name"""
        return self.strftime("%A")

    @classmethod
    def formats(cls):
        is_valid = lambda k, v: k.startswith("FORMAT")
        formats = set(v for k, v in inspect.getmembers(cls) if is_valid(k, v))
        return formats

    @classmethod
    def parse(cls, d):
        dt = cls.parse_formats(d)

        if dt is None:
            dt = cls.parse_regex(d)

        if dt is None:
            raise ValueError("Cannot infer datetime value from {}".format(d))

        return dt

    @classmethod
    def parse_formats(cls, d):
        fs = cls.formats()
        for f in fs:
            try:
                logger.debug("Attempting to parse date: {} with format: {}".format(
                    d,
                    f,
                ))
                return cls.strptime(d, f)

            except ValueError as e:
                logger.debug(e)
                pass

    @classmethod
    def parse_regex(cls, d):
        dt = None
        if re.match(r"^\-?\d+\.\d+$", d):
            # account for unix timestamps with microseconds
            logger.debug("Date: {} parsed with integer unix timestamp regex".format(d))
            dt = cls.fromtimestamp(float(d))

        elif re.match(r"^\-?\d+$", d):
            # account for unix timestamps without microseconds
            logger.debug("Date: {} parsed with float unix timestamp regex".format(d))
            val = int(d)
            dt = cls.fromtimestamp(val)

        else:
            # ISO 8601 is not very strict with the date format and this tries to
            # capture most of that leniency, with the one exception that the
            # date must be in UTC
            # https://en.wikipedia.org/wiki/ISO_8601
            m = re.match(
                r"^(\d{4}).?(\d{2}).?(\d{2}).(\d{2}):?(\d{2}):?(\d{2})(?:\.(\d+))?Z?$",
                d
            )

            if m:
                logger.debug("Date: {} parsed with ISO regex".format(d))

                parsed_dateparts = m.groups()
                dateparts = list(map(lambda x: int(x) if x else 0, parsed_dateparts[:6]))
                dt = cls(*dateparts)

                # account for ms with leading zeros
                if parsed_dateparts[6]:
                    ms_len = len(parsed_dateparts[6])
                    if ms_len >= 3:
                        millis = parsed_dateparts[6][:3]
                        micros = parsed_dateparts[6][3:] or 0

                    else:
                        millis = parsed_dateparts[6] or 0
                        micros = 0

                    # make sure each part is 3 digits by zero padding on the right
                    if millis:
                        millis = "{:0<3.3}".format(millis)
                    if micros:
                        micros = "{:0<3.3}".format(micros)

                    dt += datetime.timedelta(
                        milliseconds=int(millis),
                        microseconds=int(micros)
                    )

        return dt

    def __new__(cls, *args, **kwargs):
        if not args and not kwargs:
            return cls.utcnow()

        elif len(args) == 1 and not kwargs:
            if isinstance(args[0], datetime.datetime):
                return super(Datetime, cls).__new__(
                    cls,
                    args[0].year,
                    args[0].month,
                    args[0].day,
                    args[0].hour,
                    args[0].minute,
                    args[0].second,
                    args[0].microsecond,
                )

            elif isinstance(args[0], datetime.date):
                return super(Datetime, cls).__new__(
                    cls,
                    args[0].year,
                    args[0].month,
                    args[0].day,
                    0,
                    0,
                    0,
                    0,
                )

            elif isinstance(args[0], datetime.timedelta):
                return cls.utcnow() + args[0]

            elif isinstance(args[0], (int, float)):
                return cls.utcfromtimestamp(args[0])

            else:
                if args[0]:
                    try:
                        # if the object is pickled we would get the pickled string
                        # as our one passed in value
                        return super(Datetime, cls).__new__(cls, *args, **kwargs)

                    except TypeError:
                        fs = cls.parse(args[0])
                        if fs is None:
                            raise

                        else:
                            return fs

                else:
                    return cls.utcnow()

        else:
            return super(Datetime, cls).__new__(cls, *args, **kwargs)

    def __str__(self):
        if self.has_time():
            if self.microsecond == 0:
                return self.strftime(self.FORMAT_PRECISION_SECONDS)

            else:
                return self.strftime(self.FORMAT_ISO_8601)

        else:
            return self.strftime(self.FORMAT_PRECISION_DAY)

    def __add__(self, other):
        return type(self)(super(Datetime, self).__add__(other))

    def __sub__(self, other):
        if isinstance(other, datetime.timedelta):
            return type(self)(super(Datetime, self).__sub__(other))
        else:
            return super(Datetime, self).__sub__(other)

    def has_time(self):
        return not (
            self.hour == 0 and
            self.minute == 0 and
            self.second == 0 and 
            self.microsecond == 0
        )

    def timestamp(self):
        """
        return the current utc timestamp of self

        http://crazytechthoughts.blogspot.com/2012/02/get-current-utc-timestamp-in-python.html

        :returns: float, the current utc timestamp with microsecond precision
        """
        # this only returns second precision, which is why we don't use it
        #now = calendar.timegm(datetime.datetime.utcnow().utctimetuple())

        # this returns microsecond precision
        # http://bugs.python.org/msg180110
        epoch = datetime.datetime(1970, 1, 1)
        return (self - epoch).total_seconds()

    def iso_date(self):
        """returns datetime as ISO-8601 string with just YYYY-MM-DD"""
        return self.strftime(self.FORMAT_PRECISION_DAY)
    iso_day = iso_date
    isoday = iso_date
    isodate = iso_date

    def iso_seconds(self):
        """returns datetime as ISO-8601 string with no milliseconds"""
        return self.isoformat(timespec="seconds")
    isoseconds = iso_seconds

    def iso_8601(self):
        """returns datetime as a full ISO-8601 string with milliseconds"""
        return self.isoformat(timespec="microseconds")
    iso8691 = iso_8601

    def isofull(self):
        return self.isoformat(timespec="microseconds")

    def isoformat(self, sep="T", timespec="auto"):
        """provides python 3 compatible isoformat even for py2

        https://docs.python.org/3/library/datetime.html#datetime.datetime.isoformat

        Differences between this and builtin datetime.isoformat:

            * this will always include a timezone, defaulting to Z if no other 
            timezone is present
            * This will return YYYY-MM-DD if no time information is set and 
            timespec is "auto"

        :returns: string, the iso8604 date
        """
        add_tz = True
        if is_py2:
            if timespec == "auto":
                if self.has_time():
                    if self.microsecond == 0:
                        ret = self.strftime(self.FORMAT_PRECISION_SECONDS)

                    else:
                        ret = self.strftime(self.FORMAT_ISO_8601)

                else:
                    add_tz = False
                    ret = self.strftime(self.FORMAT_PRECISION_DAY)

            elif timespec == "hours":
                ret = self.strftime("%Y-%m-%d{}%H".format(sep))

            elif timespec == "minutes":
                ret = self.strftime("%Y-%m-%d{}%H:%M".format(sep))

            elif timespec == "seconds":
                ret = self.strftime("%Y-%m-%d{}%H:%M:%S".format(sep))

            elif timespec == "milliseconds":
                ret = self.strftime("%Y-%m-%d{}%H:%M:%S".format(sep))
                if self.microsecond == 0:
                    ret += ".000"

                else:
                    ret += ".{}".format(str(self.microsecond)[:3])

            elif timespec == "microseconds":
                ret = self.strftime("%Y-%m-%d{}%H:%M:%S.%f".format(sep))

        else:
            if timespec == "auto":
                if self.has_time():
                    ret = super(Datetime, self).isoformat(sep, timespec)
                else:
                    add_tz = False
                    ret = self.strftime(self.FORMAT_PRECISION_DAY)

            else:
                ret = super(Datetime, self).isoformat(sep, timespec)

        if add_tz and not self.tzinfo:
            if not ret.endswith("Z"):
                ret += "Z"

        return ret

    def within(self, start, stop, now=None):
        """return True if this datetime is within start and stop dates

        :param start: int|timedelta|datetime, if int, then seconds from now, so a
            negative integer would be in the past from now
        :param stop: int|timedelta|datetime, same as start
        :param now: datetime, what you want now to be
        :returns: boolean, True if self is between start and stop
        """
        if not now:
            now = type(self)()

        if isinstance(start, int):
            start = now + datetime.timedelta(seconds=start)
        start = type(self)(start)

        if isinstance(stop, int):
            stop = now + datetime.timedelta(seconds=stop)
        stop = type(self)(stop)

        return start <= self <= stop

    def strftime(self, *args, **kwargs):
        """Make sure strftime always returns a unicode object"""
        return String(super(Datetime, self).strftime(*args, **kwargs))

    def __pout__(self):
        """This just makes the object easier to digest in pout.v() calls

        more information on what pout is: https://github.com/Jaymon/pout
        """
        return self.__str__()




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import base64
import hashlib
import re
import string
import binascii
import unicodedata

from . import environ
from .compat import *


class ByteString(Bytes):
    """Wrapper around a byte string b"" to make sure we have a byte string that
    will work across python versions and handle the most annoying encoding issues
    automatically

    We treat integers like how py2.7 treats them because this is a Byte STRING and
    not just bytes, so it makes sense to return b'10' instead of 10 \x00 bytes

    :Example:
        # python 3
        s = ByteString("foo)
        str(s) # calls __str__ and returns self.unicode()
        unicode(s) # errors out
        bytes(s) # calls __bytes__ and returns ByteString
        # python 2
        s = ByteString("foo)
        str(s) # calls __str__ and returns ByteString
        unicode(s) # calls __unicode__ and returns String
        bytes(s) # calls __str__ and returns ByteString
    """
    def __new__(cls, val=b"", encoding="", errors=""):
        """
        :param val: mixed, the value you are casting to bytes
        :param encoding: string, the string encoding to use to encode/decode
        :param errors: string, how to handle errors, built-in values are:
            strict,
            ignore,
            replace,
            xmlcharrefreplace,
            backslashreplace
        """
        if isinstance(val, type(None)):
            # we do 3+ functionality even in 2.7
            if is_py2:
                raise TypeError("cannot convert 'NoneType' object to bytes")
            else:
                val = Bytes(val)

        if not encoding:
            encoding = environ.ENCODING

        if not errors:
            errors = environ.ENCODING_ERRORS

        if not isinstance(val, (bytes, bytearray)):
            if is_py2:
                val = unicode(val)
            else:
                val = str(val)
            #val = val.__str__()
            val = bytearray(val, encoding)

        instance = super(ByteString, cls).__new__(cls, val)
        instance.encoding = encoding
        instance.errors = errors
        return instance

    def __str__(self):
        return self if is_py2 else self.unicode()

    def unicode(self):
        s = self.decode(self.encoding, self.errors)
        return String(s, self.encoding, self.errors)
    __unicode__ = unicode

    def bytes(self):
        return self
    __bytes__ = bytes

    def raw(self):
        """because sometimes you need a vanilla bytes()"""
        return b"" + self

    def md5(self):
        """32 character md5 hash of string"""
        # http://stackoverflow.com/a/5297483/5006
        return hashlib.md5(self).hexdigest()

    def sha256(self):
        """64 character sh256 hash of the string"""
        return hashlib.sha256(self).hexdigest()


class String(Str):
    """Wrapper around a unicode string "" to make sure we have a unicode string that
    will work across python versions and handle the most annoying encoding issues
    automatically
    :Example:
        # python 3
        s = String("foo)
        str(s) # calls __str__ and returns String
        unicode(s) # errors out
        bytes(s) # calls __bytes__ and returns ByteString
        # python 2
        s = String("foo)
        str(s) # calls __str__ and returns ByteString
        unicode(s) # calls __unicode__ and returns String
        bytes(s) # calls __str__ and returns ByteString

    https://en.wikipedia.org/wiki/Base64
    """

    # constants from string module https://docs.python.org/3/library/string.html
    ASCII_LETTERS = string.ascii_letters
    ASCII_LOWERCASE = string.ascii_lowercase
    ASCII_UPPERCASE = string.ascii_uppercase
    DIGITS = string.digits
    HEXDIGITS = string.hexdigits
    OCTDIGITS = string.octdigits
    PUNCTUATION = string.punctuation
    PRINTABLE = string.printable
    WHITESPACE = string.whitespace

    def __new__(cls, val="", encoding="", errors=""):
        """
        :param val: mixed, the value you are casting to a string
        :param encoding: string, the string encoding to use to encode/decode
        :param errors: string, how to handle errors, built-in values are:
            strict,
            ignore,
            replace,
            xmlcharrefreplace,
            backslashreplace
        """
        if isinstance(val, type(None)):
            val = Str(val)

        if not encoding:
            encoding = environ.ENCODING

        if not errors:
            errors = environ.ENCODING_ERRORS

        if not isinstance(val, (Str, int)):
            val = ByteString(val, encoding, errors).unicode()

        instance = super(String, cls).__new__(cls, val)
        instance.encoding = encoding
        instance.errors = errors
        return instance

    def __str__(self):
        return self.bytes() if is_py2 else self

    def unicode(self):
        return self
    __unicode__ = unicode

    def bytes(self):
        s = self.encode(self.encoding)
        return ByteString(s, self.encoding, self.errors)
    __bytes__ = bytes

    def raw(self):
        """because sometimes you need a vanilla str() (or unicode() in py2)"""
        return "" + self

    def md5(self):
        """32 character md5 hash of string"""
        # http://stackoverflow.com/a/5297483/5006
        return hashlib.md5(self.bytes()).hexdigest()

    def sha256(self):
        """64 character sh256 hash of the string"""
        return hashlib.sha256(self.bytes()).hexdigest()

    def hash(self, key, name="sha256", nonce="", rounds=100000):
        """hash self with key and return the 64 byte hash

        This will produce the same hash if given the same key, it is designed to
        hash values with a dedicated key (password) and always produce the same
        hashed value if the same key (password) is always used.

        IMPORTANT In order to have the same value the same value, key, name, and
        nonce needs to be used, if you change any of these values then the hashes
        will no longer be equivalent

        :param key: string, the key/salt/password for the hash
        :param name: string, the hash to use, not required
        :param nonce: string, the nonce to use for the value, not required
        :param rounds: int, the number of rounds to hash
        :returns: string, 64 byte hex string
        """
        nonce = ByteString(nonce) if nonce else b""

        # do the actual hashing
        h = hashlib.pbkdf2_hmac(
            name,
            nonce + self.bytes(),
            String(key).bytes(), # we cast to string first, then bytes, in case key is an int
            rounds
        )
        r = binascii.hexlify(h) # convert hash to easier to consume hex
        return String(r)

    def truncate(self, size, postfix='...'):
        """similar to a normal string slice but it actually will split on a word boundary

        :Example:
            s = "foo barche"
            print s[0:5] # "foo b"
            s2 = String(s)
            print s2.truncate(5) # "foo"

        truncate a string by word breaks instead of just length
        this will guarrantee that the string is not longer than length, but it could be shorter

        http://stackoverflow.com/questions/250357/smart-truncate-in-python/250373#250373

        This was originally a method called word_truncate by Cahlan Sharp for Undrip

        :param size: int, the size you want to truncate to at max
        :param postfix: string, what you would like to be appended to the truncated
            string
        :returns: string, a new string, truncated
        """
        if len(self) < size: return self

        # our algo is pretty easy here, it truncates the string to size - postfix size
        # then right splits the string on any whitespace for a maximum of one time
        # and returns the first item of that split right stripped of whitespace
        # (just in case)
        postfix = type(self)(postfix)
        ret = self[0:size - len(postfix)]
        # if rsplit sep is None, any whitespace string is a separator
        ret = ret[:-1].rsplit(None, 1)[0].rstrip()
        return type(self)(ret + postfix)

    def indent(self, indent, count=1):
        """add whitespace to the beginning of each line of val

        http://code.activestate.com/recipes/66055-changing-the-indentation-of-a-multi-line-string/

        :param indent: string, what you want the prefix of each line to be
        :param count: int, how many times to apply indent to each line
        :returns: string, string with prefix at the beginning of each line
        """
        if not indent: return self

        s = ((indent * count) + line for line in self.splitlines(False))
        s = "\n".join(s)
        return type(self)(s)

    def stripall(self, chars):
        """Similar to the builtin .strip() but will strip chars from anywhere in the
        string

        :Example:
            s = "foo bar che.  "
            s2 = s.stripall(" .")
            print(s2) # "foobarche"
        """
        ret = ""
        for ch in self:
            if ch not in chars:
                ret += ch
        return ret

    def astrip(self, chars):
        return self.stripall(chars)

    def re(self, pattern, flags=0):
        """Provides a fluid regex interface

        :Example:
            s = 'foo##bar##che'
            parts = s.re(r'#+').split()
            print(parts) # ['foo', 'bar', 'che']
        """
        return Regex(pattern, self, flags)

    def regex(self, pattern, flags=0):
        return self.re(pattern, flags)

    def ispunc(self):
        """Returns True if all characters in string are punctuation characters"""
        for ch in self:
            if ch not in string.punctuation:
                return False
        return True

    def ispunctuation(self):
        return self.ispunc()

    def capwords(self, sep=None):
        """passthrough for string module capwords

        https://docs.python.org/3/library/string.html#string.capwords
        """
        return String(string.capwords(self, sep=sep))

    def isascii(self):
        """Return True if the string is empty or all characters in the string are
        ASCII, False otherwise.

        ASCII characters have code points in the range U+0000-U+007F.

        https://docs.python.org/3/library/stdtypes.html#str.isascii
        """
        try:
            ret = super(String, self).isascii()

        except AttributeError:
            ret = True
            a = Ascii()
            for ch in self:
                if ch not in a:
                    ret = False
                    break

        return ret

    def tokenize(self, delims=None, tokenizer_class=None):
        """Wraps the tokenizer functionality to easily tokenize a string

        :param delims: same as token.Tokenizer delims argument
        :param tokenizer_class: Tokenizer, custom class for customized functionality
        :returns: generator of String instance, each token that will also have a 
            .token that contains the raw token
        """
        if not tokenizer_class:
            from .token import Tokenizer # avoid circular dep
            tokenizer_class = Tokenizer

        tokenizer = tokenizer_class(self, delims)
        for t in tokenizer:
            st = String(t.token.text)
            st.token = t
            yield st


class Character(String):
    """Represents a unicode character, a unicode character is a set of unicode
    codepoints

    :Example:
        ch = Character("A")
        ch.hex # "0041"
    """
    @property
    def hexes(self):
        return [cp.hex for cp in self.codepoints]

    @property
    def hex(self):
        return self.unified("")

    @classmethod
    def normalize(cls, v, encoding, errors):
        codepoints = []

        if isinstance(v, bytes):
            v = v.decode(encoding, errors)

        if isinstance(v, basestring):
            if re.search(r"&[a-zA-Z0-9]{2,};?", v):
                # https://en.wikipedia.org/wiki/Character_encodings_in_HTML
                # we have a string that contains something like &ZZZZ;
                v = html.unescape(v)

            if re.match(r"^(?:[uU]\+)?[a-fA-F0-9]{4,}(?:[\s_-](?:[uU]\+)?[a-fA-F0-9]{4,})*$", v):
                # split strings like ZZZZ-ZZZZ-ZZZZ or U+ZZZZ u+zzzz etc.
                codepoints = re.split(r"[\s_-]+", v)

            else:
                # we have actual unicode codepoints (eg, sequence of \UZZZZZZZZ or \uZZZZ)
                codepoints = v

        elif isinstance(v, (list, tuple, Sequence)):
            for cp in v:
                codepoints.extend(cls.normalize(cp, encoding, errors))

        else:
            codepoints = [v]

        return codepoints

    def __new__(cls, v, encoding="", errors=""):
        if not encoding:
            encoding = environ.ENCODING

        if not errors:
            errors = environ.ENCODING_ERRORS

        codepoints = [Codepoint(cp, encoding, errors) for cp in cls.normalize(v, encoding, errors)]

        instance = super(Character, cls).__new__(cls, "".join(codepoints), encoding, errors)
        instance.codepoints = codepoints
        return instance

    def unified(self, separator=" "):
        """return the unified codepoints joined by separator

        this is just a handy way to get the codepoints with some separator, name
        comes from emojitracker

        https://medium.com/@mroth/how-i-built-emojitracker-179cfd8238ac

        :param separator: string, usually space or dash
        :returns: string, the codepoints joined by separator
        """
        cps = []
        for cp in self.codepoints:
            cps.append(cp.hex)
        return separator.join(cps)
    ligature = unified

    def graphemes(self):
        return self.codepoints

    def names(self):
        names = []
        for cp in self.codepoints:
            names.append(unicodedata.name(cp, 'UNKNOWN'))
        return names

    def __repr__(self):
        return "{} ({})".format(self, self.hex)

    def integers(self):
        """Return the codepoints as integers"""
        return list(map(int, self.codepoints))

    def __iter__(self):
        for cp in self.codepoints:
            yield cp

    def repr_string(self):
        ret = ""
        for cp in self.codepoints:
            if len(cp.hex) <= 4:
                ret += '\\u{}'.format(cp.hex)
            else:
                ret += '\\U{}'.format(cp.hex)
        return ret

    def repr_bytes(self):
        ret = ""
        if is_py2:
            for cp in self.encode(self.encoding):
                ret += "\\x{:0>2x}".format(ord(cp))

        else:
            # and to think I thought I was starting to understand unicode!
            # https://stackoverflow.com/a/54549874/5006
            s = self.encode("utf-16", "surrogatepass").decode("utf-16")
            for cp in s.encode(self.encoding):
                ret += "\\x{:0>2x}".format(cp)

        return ret

    def width(self):
        return sum(cp.width() for cp in self.codepoints)

    def __len__(self):
        return self.width()

    def eawidth(self):
        """Return the east asian width of the character

        In a broad sense, wide characters include W, F, and A (when in East Asian context),
        and narrow characters include N, Na, H, and A (when not in East Asian context)

        http://www.unicode.org/reports/tr11/tr11-36.html
        http://www.unicode.org/Public/UNIDATA/EastAsianWidth.txt

        :returns: string, a wide character: W, F, A or narrow char: N, Na, H, A
        """
        # https://docs.python.org/3/library/unicodedata.html#unicodedata.east_asian_width
        return unicodedata.east_asian_width(self)


class Codepoint(String):
    """Grapheme - a grapheme is the smallest unit of a writing system of any given language

    https://en.wikipedia.org/wiki/Grapheme
    https://docs.python.org/3/howto/unicode.html
    """
    def __new__(cls, codepoint, encoding="", errors=""):

        if not encoding:
            encoding = environ.ENCODING

        if not errors:
            errors = environ.ENCODING_ERRORS

        if isinstance(codepoint, bytes):
            codepoint = codepoint.decode(encoding, errors)

        elif isinstance(codepoint, int):
            if codepoint < 256:
                codepoint = chr(codepoint)

            else:
                codepoint = '{:X}'.format(codepoint)

        if "&" in codepoint:
            # we have a string like &ZZZZ;
            codepoint = html.unescape(codepoint)

        elif re.match(r"^[Uu]\+[a-fA-F0-9]+$", codepoint):
            # we have a string like U+ZZZZ or u+ZZZZ
            codepoint = codepoint.replace("U+", "").replace("u+", "")

        if len(codepoint) == 1:
            s = codepoint
            h = "{:0>4X}".format(ord(codepoint))

        elif len(codepoint) <= 4:
            s = '\\u{:0>4X}'.format(int(codepoint, 16)).encode(encoding).decode('unicode-escape')
            h = '{:0>4X}'.format(int(codepoint, 16)).upper()

        elif len(codepoint) <= 8:
            s = "\\U{:0>8x}".format(int(codepoint, 16)).encode(encoding).decode('unicode-escape')
            h = '{:0>4x}'.format(int(codepoint, 16)).upper()

        else:
            raise ValueError("codepoint is bigger than 8")

        instance = super(Codepoint, cls).__new__(cls, s)
        instance.hex = h
        return instance

    def __int__(self):
        return int(self.hex, 16)

    def is_wide(self):
        """Is this a wide unicode codepoint? A wide ucs4 codepoint is greater than
        4 hex digits"""
        return not self.is_narrow()

    def is_narrow(self):
        """is this a narrow unicode codepoint? A narrow ucs2 codepoint is 4 or less
        hex digits"""
        return int(self) <= 65535

    def repr_string(self):
        ret = ""
        if len(self.hex) <= 4:
            ret += '\\u{}'.format(self.hex)
        else:
            ret += '\\U{}'.format(self.hex)
        return ret

    def repr_bytes(self):
        ret = ""
        for cp in self.encode(self.encoding):
            ret += "\\x{:0>2x}".format(ord(cp))
        return ret

    def width(self):
        """A naive simple unicode width method based off of:

            https://www.cl.cam.ac.uk/~mgk25/ucs/wcwidth.c

        if you need a more intelligent solution, this SO has suggestions:

            https://stackoverflow.com/questions/30881811/how-do-you-get-the-display-width-of-combined-unicode-characters-in-python-3

        This is a really good summary of the rules for determining width:

            https://stackoverflow.com/a/9145712/5006

        search:
            * python find out width of unicode characters
            * get width of unicode character

        :returns: int, the width of the character
        """
        ucs = int(self)

        combining = [
            [ 0x0300, 0x036F ], [ 0x0483, 0x0486 ], [ 0x0488, 0x0489 ],
            [ 0x0591, 0x05BD ], [ 0x05BF, 0x05BF ], [ 0x05C1, 0x05C2 ],
            [ 0x05C4, 0x05C5 ], [ 0x05C7, 0x05C7 ], [ 0x0600, 0x0603 ],
            [ 0x0610, 0x0615 ], [ 0x064B, 0x065E ], [ 0x0670, 0x0670 ],
            [ 0x06D6, 0x06E4 ], [ 0x06E7, 0x06E8 ], [ 0x06EA, 0x06ED ],
            [ 0x070F, 0x070F ], [ 0x0711, 0x0711 ], [ 0x0730, 0x074A ],
            [ 0x07A6, 0x07B0 ], [ 0x07EB, 0x07F3 ], [ 0x0901, 0x0902 ],
            [ 0x093C, 0x093C ], [ 0x0941, 0x0948 ], [ 0x094D, 0x094D ],
            [ 0x0951, 0x0954 ], [ 0x0962, 0x0963 ], [ 0x0981, 0x0981 ],
            [ 0x09BC, 0x09BC ], [ 0x09C1, 0x09C4 ], [ 0x09CD, 0x09CD ],
            [ 0x09E2, 0x09E3 ], [ 0x0A01, 0x0A02 ], [ 0x0A3C, 0x0A3C ],
            [ 0x0A41, 0x0A42 ], [ 0x0A47, 0x0A48 ], [ 0x0A4B, 0x0A4D ],
            [ 0x0A70, 0x0A71 ], [ 0x0A81, 0x0A82 ], [ 0x0ABC, 0x0ABC ],
            [ 0x0AC1, 0x0AC5 ], [ 0x0AC7, 0x0AC8 ], [ 0x0ACD, 0x0ACD ],
            [ 0x0AE2, 0x0AE3 ], [ 0x0B01, 0x0B01 ], [ 0x0B3C, 0x0B3C ],
            [ 0x0B3F, 0x0B3F ], [ 0x0B41, 0x0B43 ], [ 0x0B4D, 0x0B4D ],
            [ 0x0B56, 0x0B56 ], [ 0x0B82, 0x0B82 ], [ 0x0BC0, 0x0BC0 ],
            [ 0x0BCD, 0x0BCD ], [ 0x0C3E, 0x0C40 ], [ 0x0C46, 0x0C48 ],
            [ 0x0C4A, 0x0C4D ], [ 0x0C55, 0x0C56 ], [ 0x0CBC, 0x0CBC ],
            [ 0x0CBF, 0x0CBF ], [ 0x0CC6, 0x0CC6 ], [ 0x0CCC, 0x0CCD ],
            [ 0x0CE2, 0x0CE3 ], [ 0x0D41, 0x0D43 ], [ 0x0D4D, 0x0D4D ],
            [ 0x0DCA, 0x0DCA ], [ 0x0DD2, 0x0DD4 ], [ 0x0DD6, 0x0DD6 ],
            [ 0x0E31, 0x0E31 ], [ 0x0E34, 0x0E3A ], [ 0x0E47, 0x0E4E ],
            [ 0x0EB1, 0x0EB1 ], [ 0x0EB4, 0x0EB9 ], [ 0x0EBB, 0x0EBC ],
            [ 0x0EC8, 0x0ECD ], [ 0x0F18, 0x0F19 ], [ 0x0F35, 0x0F35 ],
            [ 0x0F37, 0x0F37 ], [ 0x0F39, 0x0F39 ], [ 0x0F71, 0x0F7E ],
            [ 0x0F80, 0x0F84 ], [ 0x0F86, 0x0F87 ], [ 0x0F90, 0x0F97 ],
            [ 0x0F99, 0x0FBC ], [ 0x0FC6, 0x0FC6 ], [ 0x102D, 0x1030 ],
            [ 0x1032, 0x1032 ], [ 0x1036, 0x1037 ], [ 0x1039, 0x1039 ],
            [ 0x1058, 0x1059 ], [ 0x1160, 0x11FF ], [ 0x135F, 0x135F ],
            [ 0x1712, 0x1714 ], [ 0x1732, 0x1734 ], [ 0x1752, 0x1753 ],
            [ 0x1772, 0x1773 ], [ 0x17B4, 0x17B5 ], [ 0x17B7, 0x17BD ],
            [ 0x17C6, 0x17C6 ], [ 0x17C9, 0x17D3 ], [ 0x17DD, 0x17DD ],
            [ 0x180B, 0x180D ], [ 0x18A9, 0x18A9 ], [ 0x1920, 0x1922 ],
            [ 0x1927, 0x1928 ], [ 0x1932, 0x1932 ], [ 0x1939, 0x193B ],
            [ 0x1A17, 0x1A18 ], [ 0x1B00, 0x1B03 ], [ 0x1B34, 0x1B34 ],
            [ 0x1B36, 0x1B3A ], [ 0x1B3C, 0x1B3C ], [ 0x1B42, 0x1B42 ],
            [ 0x1B6B, 0x1B73 ], [ 0x1DC0, 0x1DCA ], [ 0x1DFE, 0x1DFF ],
            [ 0x200B, 0x200F ], [ 0x202A, 0x202E ], [ 0x2060, 0x2063 ],
            [ 0x206A, 0x206F ], [ 0x20D0, 0x20EF ], [ 0x302A, 0x302F ],
            [ 0x3099, 0x309A ], [ 0xA806, 0xA806 ], [ 0xA80B, 0xA80B ],
            [ 0xA825, 0xA826 ], [ 0xFB1E, 0xFB1E ], [ 0xFE00, 0xFE0F ],
            [ 0xFE20, 0xFE23 ], [ 0xFEFF, 0xFEFF ], [ 0xFFF9, 0xFFFB ],
            [ 0x10A01, 0x10A03 ], [ 0x10A05, 0x10A06 ], [ 0x10A0C, 0x10A0F ],
            [ 0x10A38, 0x10A3A ], [ 0x10A3F, 0x10A3F ], [ 0x1D167, 0x1D169 ],
            [ 0x1D173, 0x1D182 ], [ 0x1D185, 0x1D18B ], [ 0x1D1AA, 0x1D1AD ],
            [ 0x1D242, 0x1D244 ], [ 0xE0001, 0xE0001 ], [ 0xE0020, 0xE007F ],
            [ 0xE0100, 0xE01EF ]
        ]

        # test for 8-bit control characters
        if ucs == 0:
            return 0

        if (ucs < 32) or (ucs >= 0x7f and ucs < 0xa0):
            return -1

        # search in table of non-spacing characters
        for start, stop in combining:
            if start <= ucs and ucs <= stop:
                return 0

        # if we arrive here, ucs is not a combining or C0/C1 control character

        return 1 + (ucs >= 0x1100 and
            (ucs <= 0x115f or                    # Hangul Jamo init. consonants
            ucs == 0x2329 or ucs == 0x232a or
            (ucs >= 0x2e80 and ucs <= 0xa4cf and
            ucs != 0x303f) or                  # CJK ... Yi
            (ucs >= 0xac00 and ucs <= 0xd7a3) or # Hangul Syllables
            (ucs >= 0xf900 and ucs <= 0xfaff) or # CJK Compatibility Ideographs
            (ucs >= 0xfe10 and ucs <= 0xfe19) or # Vertical forms
            (ucs >= 0xfe30 and ucs <= 0xfe6f) or # CJK Compatibility Forms
            (ucs >= 0xff00 and ucs <= 0xff60) or # Fullwidth Forms
            (ucs >= 0xffe0 and ucs <= 0xffe6) or
            (ucs >= 0x20000 and ucs <= 0x2fffd) or
            (ucs >= 0x30000 and ucs <= 0x3fffd)))


class Regex(object):
    """Provides a passthrough interface to the re module to run pattern against s

    https://docs.python.org/3/library/re.html
    """
    def __init__(self, pattern, s, flags=0):
        self.pattern = pattern
        self.s = s
        self.flags = flags

    def search(self, flags=0):
        flags |= self.flags
        return re.search(self.pattern, self.s, flags)

    def match(self, flags=0):
        flags |= self.flags
        return re.match(self.pattern, self.s, flags)

    def fullmatch(self, flags=0):
        flags |= self.flags
        return re.fullmatch(self.pattern, self.s, flags)

    def split(self, maxsplit=0, flags=0):
        flags |= self.flags
        return re.split(self.pattern, self.s, maxsplit=maxsplit, flags=flags)

    def findall(self, flags=0):
        flags |= self.flags
        return re.findall(self.pattern, self.s, flags=flags)

    def finditer(self, flags=0):
        flags |= self.flags
        return re.finditer(self.pattern, self.s, flags=flags)

    def sub(self, repl, count=0, flags=0):
        flags |= self.flags
        return re.sub(self.pattern, repl, self.s, count=count, flags=flags)

    def subn(self, repl, count=0, flags=0):
        flags |= self.flags
        return re.subn(self.pattern, repl, self.s, count=count, flags=flags)


class Base64(String):
    """This exists to normalize base64 encoding between py2 and py3, it assures that
    you always get back a unicode string when you encode or decode and that you can
    pass in a unicode or byte string and it just works
    """
    @classmethod
    def encode(cls, s, encoding=""):
        """converts a plain text string to base64 encoding
        :param s: unicode str|bytes, the base64 encoded string
        :returns: unicode str
        """
        b = ByteString(s, encoding=encoding)
        be = base64.b64encode(b).strip()
        return String(be)

    @classmethod
    def decode(cls, s, encoding=""):
        """decodes a base64 string to plain text
        :param s: unicode str|bytes, the base64 encoded string
        :returns: unicode str
        """
        b = ByteString(s)
        bd = base64.b64decode(b)
        return String(bd, encoding=encoding)


class HTMLCleaner(HTMLParser):
    """strip html tags from a string

    :example:
        html = "this is <b>some html</b>
        text = HTMLCleaner.strip_tags(html)
        print(text) # this is some html

    http://stackoverflow.com/a/925630/5006
    https://docs.python.org/2/library/htmlparser.html
    """
    # https://developer.mozilla.org/en-US/docs/Web/HTML/Block-level_elements
    BLOCK_TAGNAMES = set([
        "address",
        "article",
        "aside",
        "blockquote",
        "br",
        "canvas",
        "dd",
        "div",
        "dl",
        "fieldset", 
        "figcaption",
        "figure",
        "footer",
        "form",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "header", 
        "hgroup", 
        "hr",
        "li",
        "main",
        "nav",
        "noscript",
        "ol",
        "output",
        "p",
        "pre",
        "section",
        "table",
        "tfoot",
        "ul",
        "video",
    ])

    @classmethod
    def strip_tags(cls, html, *args, **kwargs):
        s = cls(*args, **kwargs)
        # convert entities back otherwise stripper will get rid of them
        # http://stackoverflow.com/a/28827374/5006
        #html = s.unescape(html)
        s.feed(html)
        return s.get_data()

    def __init__(self, block_sep="\n", inline_sep="", keep_img_src=False):
        """create an instance and configure it

        :param block_sep: string, strip a block tag and then add this to the end of the
            stripped tag, so if you have <p>foo bar<p> and block_sep=\n, then the stripped
            string would be foo bar\n
        :param inline_sep: string, same as block_sep, but gets added to the end of the
            stripped inline tag
        :param keep_img_src: boolean, if True, the img.src attribute will replace the <img />
            tag
        """
        self.fed = []
        self.block_sep = block_sep
        self.inline_sep = inline_sep
        self.keep_img_src = keep_img_src

        if is_py2:
            HTMLParser.__init__(self)
        else:
            #self.reset()
            super(HTMLCleaner, self).__init__()

    def handle_data(self, d):
        self.fed.append(d)

    def handle_entityref(self, name):
        self.fed.append("&{};".format(name))

    def handle_starttag(self, tag, attrs):
        # https://docs.python.org/3/library/html.parser.html#html.parser.HTMLParser.handle_starttag
        if tag == "img" and self.keep_img_src:
            for attr_name, attr_val in attrs:
                if attr_name == "src":
                    self.fed.append("\n{}\n".format(attr_val))

    def handle_endtag(self, tagname):
        if tagname in self.BLOCK_TAGNAMES:
            if self.block_sep:
                self.fed.append(self.block_sep)
        else:
            if self.inline_sep:
                self.fed.append(self.inline_sep)

    def get_data(self):
        return self.unescape("".join(self.fed))

    @classmethod
    def unescape(cls, s):
        """unescapes html entities (eg, &gt;) to plain text (eg, &gt; becomes >)"""
        # https://stackoverflow.com/a/2087433/5006
        return html.unescape(s)


class Ascii(String):
    """This is just a convenience class so I can get all punctuation characters, 
    but I added other methods in case I need to use this somewhere else"""
    def __new__(cls):
        s = list(chr(x) for x in range(0x00, 0x7F + 1))
        instance = super(Ascii, cls).__new__(cls, s)
        return instance

    def punctuation(self):
        """Return all the ascii punctuation characters"""
        ret = []
        for x in range(0x00, 0x7F + 1): # 0-127
            if x == 96: continue
            if x == 26: continue

            s = chr(x)
            if s.isprintable() and not s.isspace() and not s.isalnum():
                ret.append(s)

        return "".join(ret)

    def alpha(self):
        """Return upper and lower case letters"""
        ret = []
        for x in range(0x00, 0x7F + 1): # 0-127
            s = chr(x)
            if s.isalpha():
            #if s.isalnum():
                ret.append(s)

        return "".join(ret)

    def num(self):
        """Return numbers 0-9"""
        ret = []
        for x in range(0x00, 0x7F + 1): # 0-127
            s = chr(x)
            if s.isnumeric():
            #if s.isalnum():
                ret.append(s)

        return "".join(ret)

    def numeric(self):
        return self.num()

    def alnum(self):
        """Return letters and numbers only"""
        ret = [c for c in self.alpha()]
        ret.update(c for c in self.num())
        return ret

    def alphanumeric(self):
        return self.alnum()

    def alphanum(self):
        return self.alnum()




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import os
import tempfile

from .compat import *


class Environ(object):
    """Create an Environ for the given modname module

    you would usually create this like this:

        environ = Environ(__name__)

    This would take __name__ and normalize it, so something like `foo.bar` would
    become `FOO_`, if you want to set the namespace explicitely, you could do:

        environ = Environ("FOO_")
    """
    def __init__(self, modname=""):
        self.namespace = ""
        if modname:
            namespace = modname.split(".")[0].upper()
            if not namespace.endswith("_"):
                namespace += "_"
            self.namespace = namespace

    def set(self, key, value):
        k = self.key(key)
        os.environ[k] = value

    def nset(self, key, values):
        """Given a list of values, this will set key_* where * is 1 -> len(values)

        :param key: string, the key that will be added to the environment as key_*
        :param values: list, a list of values
        """
        self.ndelete(key)

        for i, v in enumerate(values, 1):
            k = "{}_{}".format(key, i)
            self.set(k, v)

    def delete(self, key):
        """remove key from the environment"""
        k = self.key(key)
        os.environ.pop(k)

    def ndelete(self, key):
        """remove all key_* from the environment"""
        for k in self.nkeys(key):
            os.environ.pop(k)

    def get(self, key, default=None):
        """get a value for key from the environment

        :param key: string, this will be normalized using the key() method
        :returns: mixed, the value in the environment of key, or default if key
            is not in the environment
        """
        if self.namespace and not key.startswith(self.namespace):
            key = self.namespace + key

        r = os.environ.get(key, default)
        return r

    def nget(self, key):
        """this will look for key, and kkey_N (where
        N is 1 to infinity) in the environment

        The num checks (eg *_1, *_2) go in order, so you can't do *_1, *_3, because it
        will fail on missing *_2 and move on, so make sure your nums are in order 
        (eg, 1, 2, 3, ...)

        :param key: string, the name of the environment variables
        :returns: generator, the found values for key and key_* variants
        """
        for nkey in self.nkeys(key):
            yield os.environ[nkey]

    def keys(self):
        """yields all the keys of the given namespace currently in the environment"""
        for k in os.environ.keys():
            if k.startswith(self.namespace):
                yield k

    def key(self, key):
        """normalizes key to have the namespace

        :Example:
            environ = Environ("FOO_")
            k = environ.key("BAR")
            print(k) # FOO_BAR
        """
        if self.namespace and not key.startswith(self.namespace):
            key = self.namespace + key
        return key

    def nkey(self, key, n):
        """helper method for nkeys"""
        k = self.key(key)
        for fmt in ["{key}_{n}", "{key}{n}"]:
            nkey = fmt.format(key=k, n=n)
            if nkey in os.environ:
                return nkey

        raise KeyError("Environ {} has no {} key".format(key, n))

    def nkeys(self, key):
        """This returns the actual environment variable names from * -> *_N

        :param key: string, the name of the environment variables
        :returns: generator, the found environment names
        """
        k = self.key(key)
        if k in os.environ:
            yield k

        # now try importing _1 -> _N prefixes
        n = 0
        while True:
            try:
                yield self.nkey(key, n)

            except KeyError:
                # 0 is a special case, so if it fails we keep going
                if n:
                    # we are done because we missed an n value
                    break

            finally:
                n += 1

    def paths(self, key):
        """Similar to .all() but splits each value using the path separator"""
        sep = os.pathsep
        for paths in self.nget(key):
            for p in paths.split(sep):
                yield p


environ = Environ(modname=__name__)

ENCODING = environ.get("ENCODING", "UTF-8")
"""For things that need encoding, this will be the default encoding if nothing else
is passed in"""

ENCODING_ERRORS = environ.get("ENCODING_ERRORS", "replace")
"""For things that need encoding, this will handle any errors"""

# 2021-1-15 - we place our cache by default into another directory because I have problems
# running some commands like touch on the base temp directory on MacOS on
# Python3, no idea why but I couldn't get around it
CACHE_DIR = environ.get("CACHE_DIR", os.path.join(tempfile.gettempdir(), environ.namespace.lower().rstrip("_")))
"""the default caching directory for things that need a cache folder"""




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import re
import inspect
from socket import gethostname

from .compat import *
from .string import String, ByteString
from .path import Cachepath, Path


class Url(String):
    """a url object on steroids, this is here to make it easy to manipulate urls
    we try to map the supported fields to their urlparse equivalents, with some additions

    https://tools.ietf.org/html/rfc3986.html

    given a url http://user:pass@foo.com:1000/bar/che?baz=boom#anchor

    .scheme = http
    .netloc = user:pass@foo.com:1000
    .hostloc = foo.com:1000
    .hostname = foo.com
    .host() = http://foo.com
    .port = 1000
    .base = http://user:pass@foo.com:1000/bar/che
    .fragment = anchor
    .anchor = fragment
    .uri = /bar/che?baz=boom#anchor
    .host(...) = http://foo.com/...
    .base(...) = http://foo.com/bar/che/...
    """
    scheme = "http"

    username = None

    password = None

    hostname = ""

    port = None

    netloc = ""

    path = ""

    query_kwargs = {}

    fragment = ""

    @property
    def root(self):
        """just return scheme://netloc"""
        return parse.urlunsplit((
            self.scheme,
            self.netloc,
            "",
            "",
            ""
        ))

    @property
    def anchor(self):
        """alternative name for fragment"""
        return self.fragment

    @property
    def uri(self):
        """return the uri, which is everything but base (no scheme, host, etc)"""
        uristring = self.path
        if self.query:
            uristring += "?{}".format(self.query)
        if self.fragment:
            uristring += "#{}".format(self.fragment)

        return uristring

    @classmethod
    def keys(cls):
        # we need to ignore property objects also
        is_valid = lambda k, v: not k.startswith("__") and not callable(v) and not isinstance(v, property)
        keys = set(k for k, v in inspect.getmembers(cls) if is_valid(k, v))
        return keys

    @classmethod
    def default_values(cls):
        """Return the default values for the defined keys in the class

        :returns: dict, the key is the default class variable name and the value
            is the value it is set to
        """
        values = {}
        for k in cls.keys():
            v = getattr(cls, k)
            if isinstance(v, dict):
                values[k] = dict(v)

            elif isinstance(v, list):
                values[k] = list(v)

            else:
                values[k] = v

        return values

    @classmethod
    def merge(cls, urlstring="", **kwargs):
        # we handle port before any other because the port of host:port in hostname takes precedence
        # the port on the host would take precedence because proxies mean that the
        # host can be something:10000 and the port could be 9000 because 10000 is
        # being proxied to 9000 on the machine, but we want to automatically account
        # for things like that and then if custom behavior is needed then this method
        # can be overridden
        parts = cls.default_values()
#         parts = {
#             "hostname": cls.hostname,
#             "port": cls.port,
#             "query_kwargs": dict(cls.query_kwargs),
#             "scheme": cls.scheme,
#             "netloc": cls.netloc,
#             "path": cls.path,
#             "fragment": cls.fragment,
#             "username": cls.username,
#             "password": cls.password,
#         }

        if urlstring:
            properties = [
                "scheme", # 0
                "netloc", # 1
                "path", # 2
                "fragment", # 5
                "username",
                "password",
                "hostname",
                "port",
                "query", # 4
            ]

            if re.match(r"^\S+://", urlstring) or urlstring.startswith("//"):
                o = parse.urlsplit(String(urlstring))

            else:
                # if we don't have a scheme we put // in front of it so it will
                # still parse correctly
                s = "//{}".format(String(urlstring))
                o = parse.urlsplit(s)

            for k in properties:
                v = getattr(o, k)
                if v:
                    parts[k] = v

            query = parts.get("query", "")
            if query:
                parts["query_kwargs"].update(cls.parse_query(query))

        query = kwargs.pop("query", "")
        if query:
            parts["query_kwargs"].update(cls.parse_query(query))

        query_kwargs = kwargs.pop("query_kwargs", {})
        if query_kwargs:
            parts["query_kwargs"].update(query_kwargs)

        parts["query"] = ""
        if parts["query_kwargs"]:
            parts["query"] = cls.unparse_query(parts["query_kwargs"])

        for k, v in kwargs.items():
            parts[k] = v

        ports = {
            "http": 80,
            "https": 443,
        }
        domain, port = cls.split_hostname_from_port(parts["hostname"])
        parts["hostname"] = domain
        if port:
            parts["port"] = kwargs.get("port", port)

        if not parts.get("port", None):
            if "default_port" in kwargs:
                parts["port"] = kwargs["default_port"]
            else:
                parts["port"] = ports.get(parts["scheme"], None)

        # make sure port is an int
        if parts["port"]:
            parts["port"] = int(parts["port"])

        hostloc = parts["hostname"]
        port = parts["port"]
        # we don't want common ports to be a part of a .geturl() call, but we do
        # want .port to return them
        if port and port not in set(ports.values()):
            hostloc = '{}:{}'.format(hostloc, port)
        parts["hostloc"] = hostloc

        parts["netloc"] = parts["hostloc"]
        username = kwargs.get("username", parts["username"])
        password = kwargs.get("password", parts["password"])
        if username:
            parts["netloc"] = "{}:{}@{}".format(
                username or "",
                password or "",
                parts["hostloc"]
            )

        parts["path"] = "/".join(cls.normalize_paths(parts["path"]))

        parts["urlstring"] = parse.urlunsplit((
            parts["scheme"],
            parts["netloc"],
            parts["path"],
            parts["query"],
            parts["fragment"],
        ))

        for k in parts:
            if isinstance(parts[k], bytes):
                parts[k] = String(parts[k])

        return parts

    @classmethod
    def parse_query(cls, query):
        """return name=val&name2=val2 strings into {name: val} dict"""
        if not query: return {}

        if isinstance(query, bytes):
            query = String(query)

        # https://docs.python.org/2/library/urlparse.html
        query_kwargs = parse.parse_qs(query, True, strict_parsing=True)
        return cls.normalize_query_kwargs(query_kwargs)

    @classmethod
    def normalize_query_kwargs(cls, query):
        d = {}
        # https://docs.python.org/2/library/urlparse.html
        for k, kv in query.items():
            #k = k.rstrip("[]") # strip out php type array designated variables
            if isinstance(k, bytes):
                k = String(k)

            if len(kv) > 1:
                d[k] = kv

            else:
                d[k] = kv[0]

        return d

    @classmethod
    def unparse_query(cls, query_kwargs):
        return urlencode(query_kwargs, doseq=True)

    @classmethod
    def normalize_paths(cls, *paths):
        args = []
        for ps in paths:
            if isinstance(ps, basestring):
                args.extend(filter(None, ps.split("/")))
                #args.append(ps.strip("/"))
            else:
                for p in ps:
                    args.extend(cls.normalize_paths(p))
        return args

    @classmethod
    def split_hostname_from_port(cls, hostname, default_port=None):
        """given a hostname:port return a tuple (hostname, port)"""
        bits = hostname.split(":", 2)
        p = int(default_port) if default_port else default_port
        d = bits[0]
        if len(bits) == 2:
            p = int(bits[1])
        return d, p

    def __new__(cls, urlstring=None, **kwargs):
        parts = cls.merge(urlstring, **kwargs)
        urlstring = parts.pop("urlstring")
        instance = super(Url, cls).__new__(cls, urlstring)
        for k, v in parts.items():
            setattr(instance, k, v)
        return instance

    def create(self, *args, **kwargs):
        return type(self)(*args, **kwargs)

    def add(self, **kwargs):
        """Just a shortcut to change the current url, equivalent to Url(self, **kwargs)"""
        if "path" in kwargs:
            path = kwargs["path"]
            if isinstance(path, bytes):
                path = String(path)
            if not path[0].startswith("/"):
                paths = self.normalize_paths(self.path, path)
            else:
                paths = self.normalize_paths(path)
            kwargs["path"] = "/".join(paths)
        return self.create(self, **kwargs)

    def subtract(self, *paths, **kwargs):
        sub_kwargs = self.jsonable()

        path2 = self.normalize_paths(paths)
        path2.extend(self.normalize_paths(kwargs.pop("path", "")))
        if path2:
            sub_path = self.normalize_paths(self.path)
            for p in path2:
                try:
                    sub_path.remove(p)
                except ValueError:
                    pass

            sub_kwargs["path"] = sub_path

        for k, v in kwargs.items():
            if k == "query_kwargs":
                for qk, qv in kwargs[k].items():
                    if str(sub_kwargs[k][qk]) == str(qv):
                        sub_kwargs[k].pop(qk)

            else:
                if str(sub_kwargs[k]) == str(v):
                    sub_kwargs.pop(k)

        return self.create(**sub_kwargs)

    def _normalize_params(self, *paths, **query_kwargs):
        """a lot of the helper methods are very similar, this handles their arguments"""
        kwargs = {}

        if paths:
            fragment = paths[-1]
            if fragment:
                if fragment.startswith("#"):
                    kwargs["fragment"] = fragment
                    paths.pop(-1)

            kwargs["path"] = "/".join(self.normalize_paths(*paths))

        kwargs["query_kwargs"] = query_kwargs
        return kwargs

    def parent(self, *paths, **query_kwargs):
        """create a new Url instance one level up from the current Url instance

        so if self contains /foo/bar then self.parent() would return /foo

        :param *paths: list, the paths to append to the parent path
        :param **query_kwargs: dict, any query string params to add
        :returns: new Url instance
        """
        kwargs = self._normalize_params(*paths, **query_kwargs)
        path_args = self.path.split("/")
        if path_args:
            urlstring = self.subtract(path_args[-1])
        else:
            urlstring = self

        return urlstring.add(**kwargs)

    def base(self, *paths, **query_kwargs):
        """create a new url object using the current base path as a base

        if you had requested /foo/bar, then this would append *paths and **query_kwargs
        to /foo/bar

        :example:
            # current path: /foo/bar

            print url # http://host.com/foo/bar

            print url.base() # http://host.com/foo/bar
            print url.base("che", boom="bam") # http://host/foo/bar/che?boom=bam

        :param *paths: list, the paths to append to the current path without query params
        :param **query_kwargs: dict, any query string params to add
        """
        kwargs = self._normalize_params(*paths, **query_kwargs)
        if self.path:
            if "path" in kwargs:
                paths = self.normalize_paths(self.path, kwargs["path"])
                kwargs["path"] = "/".join(paths)
            else:
                kwargs["path"] = self.path
        return self.create(self.root, **kwargs)

    def host(self, *paths, **query_kwargs):
        """create a new url object using the host as a base

        if you had requested http://host/foo/bar, then this would append *paths and **query_kwargs
        to http://host

        :example:
            # current url: http://host/foo/bar

            print url # http://host.com/foo/bar

            print url.host() # http://host.com/
            print url.host("che", boom="bam") # http://host/che?boom=bam

        :param *paths: list, the paths to append to the current path without query params
        :param **query_kwargs: dict, any query string params to add
        """
        kwargs = self._normalize_params(*paths, **query_kwargs)
        return self.create(self.root, **kwargs)

    def copy(self):
        return self.__deepcopy__()

    def __copy__(self):
        return self.__deepcopy__()

    def __deepcopy__(self, memodict={}):
        return self.create(
            scheme=self.scheme,
            username=self.username,
            password=self.password,
            hostname=self.hostname,
            port=self.port,
            path=self.path,
            query_kwargs=self.query_kwargs,
            fragment=self.fragment,
        )

    def __add__(self, other):
        ret = ""
        if isinstance(other, Mapping):
            ret = self.add(query_kwargs=other)

        elif isinstance(other, MutableSequence):
            ret = self.add(path=other)

        elif isinstance(other, basestring):
            ret = self.add(path=other)

        elif isinstance(other, Sequence):
            ret = self.add(path=other)

        else:
            raise ValueError("Not sure how to add {}".format(type(other)))

        return ret
    __iadd__ = __add__

    def __truediv__(self, other):
        ret = ""
        if isinstance(other, MutableSequence):
            ret = self.add(path=other)

        elif isinstance(other, basestring):
            ret = self.add(path=other)

        elif isinstance(other, Sequence):
            ret = self.add(path=other)

        else:
            raise ValueError("Not sure how to add {}".format(type(other)))

        return ret
    __itruediv__ = __truediv__

    def __sub__(self, other):
        """Return a new url with other removed"""
        ret = ""
        if isinstance(other, Mapping):
            ret = self.subtract(query_kwargs=other)

        elif isinstance(other, MutableSequence):
            ret = self.subtract(path=other)

        elif isinstance(other, basestring):
            ret = self.subtract(path=other)

        elif isinstance(other, Sequence):
            ret = self.subtract(path=other)

        else:
            raise ValueError("Not sure how to add {}".format(type(other)))

        return ret
    __isub__ = __sub__

    def jsonable(self):
        ret = {}
        for k in self.keys():
            v = getattr(self, k)
            if k == "query_kwargs":
                ret[k] = dict(v)
            else:
                ret[k] = v

        return ret


class Host(tuple):
    """Creates a tuple (hostname, port) that can be passed to built-in PYthon server
    classes and anything that uses that same interface, does all the lifting to
    figure out what the hostname and port should be based on the passed in input,
    so you can pass things like hostname:port, or (host, port), etc.
    """
    @property
    def hostname(self):
        """Return just the hostname with no port or scheme or anything

        :returns: string
        """
        return self[0]

    @property
    def port(self):
        """return just the port

        :returns: int
        """
        return self[1]

    @property
    def hostloc(self):
        """Returns hostname:port

        :returns: string
        """
        return "{}:{}".format(self.hostname, self.port)
    netloc = hostloc

    def __new__(cls, host, port=None):
        u = Url(host, default_port=port)
        return super(Host, cls).__new__(cls, [u.hostname, u.port if u.port else 0])

    def __str__(self):
        return self.__bytes__() if is_py2 else self.__unicode__()

    def __unicode__(self):
        return String(self.hostloc)

    def __bytes__(self):
        return ByteString(self.hostloc)

    def client(self):
        """Url can technically hold a hostname like 0.0.0.0, this will compensate
        for that, useful for test clients

        :returns: a netloc (host:port) that a client can use to make a request
        """
        netloc = ""
        domain, port = self
        netloc = gethostname() if domain == "0.0.0.0" else domain
        if port:
            netloc += ":{}".format(port)
        return netloc



class Urlpath(Cachepath):

    def __new__(cls, url, path="", **kwargs):
        # TODO -- pass in client? I should move testdata.client.HTTP (and really
        # all of testdata.client into .client.py module

        keys = []
        if path:
            path = Path.create_file(path)
            kwargs.setdefault("dir", path.parent)
            keys.append(path.basename)

        instance = super(Urlpath, cls).__new__(cls, *keys, **kwargs)
        instance.url = url
        return instance


    def read(self):
        raise NotImplementedError()
        # TODO -- if the file doesn't exist fetch it then read
        pass

    def read_text(self):
        raise NotImplementedError()
        # TODO -- if the file doesn't exist fetch it then read
        pass

    def read_bytes(self):
        raise NotImplementedError()
        # TODO -- if the file doesn't exist fetch it then read
        pass




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
from io import SEEK_SET, SEEK_CUR, SEEK_END
import re
import itertools
import logging
from contextlib import contextmanager
from string import whitespace as WHITESPACE, punctuation as PUNCTUATION
import io

from .compat import *
from .string import String, ByteString


logger = logging.getLogger(__name__)


class BaseToken(object):
    """The base for the Token and SubToken containing shared functionality"""
    def __str__(self):
        return ByteString(self.__unicode__()) if is_py2 else self.__unicode__()


class SubToken(BaseToken):
    """The subtoken is one of: left-delim, token, right-delim, these are created
    and added to a Token which is returned from the Tokenizer"""
    def __init__(self, tokenizer, text, start, stop):
        """
        :param tokenizer: StreamTokenizer, the tokenizer creating this subtoken
        :param text: string, the text of this subtoken
        :param start: int, the cursor offset this subtoken starts at in the tokenizer
        :param stop: int, the cursor offset this subtoken ends at in the tokenizer
        """
        self.text = text
        self.start = start
        self.stop = stop
        self.tokenizer = tokenizer

    def __unicode__(self):
        return self.text


class Token(BaseToken):
    """This is what is returned from the Tokenizer and contains 3 subtokens:

        .ldelim - the deliminators to the left of the token
        .token - the actual token that was found
        .rdelim - the deliminators to the right of the token
    """
    def __init__(self, tokenizer, ldelim, token, rdelim):
        self.ldelim = ldelim
        self.token = token
        self.rdelim = rdelim
        self.tokenizer = tokenizer

    def __pout__(self):
        """used by pout python external library

        https://github.com/Jaymon/pout
        """
        tokens = (
            '"{}"'.format(self.ldelim.text) if self.ldelim else None,
            '"{}"'.format(self.token.text),
            '"{}"'.format(self.rdelim.text) if self.rdelim else None,
        )

        return "{}, {}, {}".format(tokens[0], tokens[1], tokens[2])

    def __unicode__(self):
        return self.token.__unicode__()


class StreamTokenizer(io.IOBase):
    """Tokenize a string finding tokens that are divided by pass in deliminators

    https://docs.python.org/3/library/io.html#io.IOBase
    """
    DEFAULT_DELIMS = WHITESPACE + PUNCTUATION
    """IF no deliminators are passed into the constructor then use these"""

    token_class = Token
    """The token class this class will use to create Token instances"""

    subtoken_class = SubToken
    """The subtoken class this class will use to create SubToken instances"""

    def __init__(self, stream, delims=None):
        """
        :param stream: io.IOBase, this is the input that will be tokenized, the stream
            has to be seekable
        :param delims: callback|string, if a callback, it should have the signature:
            callback(char) and return True if the char is a delim, False otherwise.
            If a string then it is a string of chars that will be considered delims
        """
        if not is_py2:
            # python 2 will just raise an error when we try and seek
            if not stream.seekable():
                raise ValueError("Unseekable streams are not supported")

        self.stream = stream

        if callable(delims):
            self.is_delim = delims

        else:
            if not delims:
                delims = self.DEFAULT_DELIMS

            delims = set(delims)
            self.is_delim = lambda ch: ch in delims

    def tell_ldelim(self):
        """Tell the current ldelim start position, this is mainly used internally

        :returns: int, the cursor position of the start of the left deliminator of
            the current token
        """
        pos = self.stream.tell()
        ch = self.stream.read(1)
        if not ch:
            # EOF, stream is exhausted
            raise StopIteration()

        if self.is_delim(ch):
            p = pos
            while self.is_delim(ch):
                p -= 1
                if p >= 0:
                    self.stream.seek(p)
                    ch = self.stream.read(1)

                else:
                    p = 0
                    break

            if p > 0:
                p += 1
            pos = p

        else:
            p = pos
            while not self.is_delim(ch):
                p -= 1
                if p >= 0:
                    self.stream.seek(p)
                    ch = self.stream.read(1)

                else:
                    break

            if p >= 0:
                self.stream.seek(p)
                pos = self.tell_ldelim()

            else:
                pos = -1

        return pos

    def __iter__(self):
        self.stream.seek(0)
        return self

    def peek(self):
        """Return the next token but don't increment the cursor offset"""
        ret = None
        with self.temporary() as t:
            try:
                ret = t.next()
            except StopIteration:
                pass
        return ret

    def next(self):
        """Get the next Token

        :returns: Token, the next token found in .stream
        """
        ldelim = token = rdelim = None

        start = self.tell_ldelim()

        if start >= 0:
            text = ""
            self.stream.seek(start)
            ch = self.stream.read(1)

            while self.is_delim(ch):
                text += ch
                ch = self.stream.read(1)

            stop = self.stream.tell() - 1
            ldelim = self.subtoken_class(self, text, start, stop)
            start = stop

        else:
            start = 0
            self.stream.seek(start)
            ch = self.stream.read(1)

        if ch:
            text = ""
            while ch and not self.is_delim(ch):
                text += ch
                ch = self.stream.read(1)

            stop = self.stream.tell() - 1
            token = self.subtoken_class(self, text, start, stop)
            start = stop

        if ch:
            text = ""
            while self.is_delim(ch):
                text += ch
                ch = self.stream.read(1)

            stop = self.stream.tell() - 1
            rdelim = self.subtoken_class(self, text, start, stop)

        #if not ldelim and not token and not rdelim:
        if not token:
            raise StopIteration()

        return self.token_class(self, ldelim, token, rdelim)

    def __next__(self):
        return self.next()

    def prev(self):
        """Returns the previous Token

        :returns: Token, the previous token found in the stream
        """
        token = None
        try:
            start = self.tell_ldelim()

        except StopIteration:
            self.seek(self.tell() - 1)
            start = self.tell_ldelim()
            token = self.next()

        else:
            if start > 0:
                self.seek(start - 1)
                start = self.tell_ldelim()
                token = self.next()

        if token:
            start = token.ldelim.start if token.ldelim else token.token.start
            self.seek(start)
        return token

    def read(self, count=-1):
        """Read count tokens and return them

        :param count: int, if >0 then return count tokens, if -1 then return all
            remaining tokens
        :returns: list, the read Token instances
        """
        ret = []
        if count:
            if count > 0:
                while count > 0:
                    try:
                        ret.append(self.next())
                    except StopIteration:
                        break

                    else:
                        count -= 1

            else:
                while True:
                    try:
                        ret.append(self.next())
                    except StopIteration:
                        break

        return ret

    def readall(self):
        """Read and return all remaining tokens"""
        return self.read()

    def fileno(self):
        return self.stream.fileno()

    def readable(self):
        return self.stream.readable()

    def writeable(self):
        """https://docs.python.org/3/library/io.html#io.IOBase.writable"""
        return False

    def tell(self):
        """Return the current stream position"""
        return self.stream.tell()

    def seek(self, offset, whence=SEEK_SET):
        """Change the stream position to the given byte offset. offset is interpreted
        relative to the position indicated by whence.

        The default value for whence is SEEK_SET. Values for whence are:

            * SEEK_SET or 0 – start of the stream (the default); offset should be zero or positive
            * SEEK_CUR or 1 – current stream position; offset may be negative

        Return the new absolute position.
        """
        offset = int(offset)

        if whence == SEEK_SET:
            self.stream.seek(max(0, offset))

        elif whence == SEEK_CUR:
            self.stream.seek(max(0, self.tell() + offset))

#         elif whence == SEEK_END:
#             raise NotImplemented()
#             #self.offset = max(0, len(self) - offset)

        else:
            raise ValueError("Unknown or unsupported whence value: {}".format(whence))

        return self.tell()

    def seekable(self):
        """https://docs.python.org/3/library/io.html#io.IOBase.seekable"""
        return self.stream.seekable()

    @contextmanager
    def transaction(self):
        """If an error is raised reset the cursor back to where the transaction
        was started"""
        start = self.tell()
        try:
            yield self

        except Exception as e:
            self.seek(start)
            raise

    @contextmanager
    def temporary(self):
        """similar to .transaction() but will always discard anything read and
        reset the cursor back to where it started, you use this because you want
        to check some tokens ephemerally"""
        start = self.tell()
        try:
            yield self

        finally:
            self.seek(start)

    def close(self, *args, **kwargs):
        raise NotImplementedError()

    def closed(self, *args, **kwargs):
        raise NotImplementedError()

    def flush(self, *args, **kwargs):
        raise NotImplementedError()

    def readline(self, size=-1):
        raise NotImplementedError()

    def readline(self, hint=-1):
        raise NotImplementedError()


class Tokenizer(StreamTokenizer):
    """Extends stream tokenizer to accept strings or IO streams"""
    def __init__(self, mixed, delims=None):
        if isinstance(mixed, basestring):
            mixed = StringIO(String(mixed))

        super(Tokenizer, self).__init__(mixed, delims)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import

from .compat import *
from .string import String


class EnumMixin(object):
    @classmethod
    def find_name(cls, name):
        """find the name of an enum, case-insensitive

        :param name: str|int, can be anything and it will be converted to an int
        :returns: string, the value found for name
        """
        if isinstance(name, int):
            ret = cls.convert_int_to_str(name)
        else:
            cls.convert_str_to_int(name) # error check
            ret = cls.convert_str_to_str(name)
        return String(ret)

    @classmethod
    def find_value(cls, value):
        """find the value of an enum, similar to .find_name() this will normalize
        the name or value and return the found value

        :param value: str|int, an enum value or name
        :returns value: int, the value of the enum
        """
        if isinstance(value, int):
            cls.convert_int_to_str(value) # error check
            ret = value

        else:
            ret = cls.convert_str_to_int(value)
        return ret

    @classmethod
    def all(cls):
        """return all the enum values with names as key and integer values"""
        d = {}
        for k, v in cls.__members__.items():
            d[String(k)] = v.value
        return d

    @classmethod
    def names(cls):
        """return all the ENUM names"""
        return cls.all().keys()

    @classmethod
    def values(cls):
        """return all the enum values"""
        return cls.all().values()

    @classmethod
    def convert_str_to_int(cls, name):
        name = cls.convert_str_to_str(name)
        value = getattr(cls, name)
        return value.value

    @classmethod
    def convert_int_to_str(cls, value):
        found = False
        for k, v in cls.__members__.items():
            if v.value == value:
                ret = k
                found = True
                break
        if not found:
            raise ValueError("Value {} is not a valid enumerated value".format(value))

        return ret

    @classmethod
    def convert_str_to_str(cls, name):
        if isinstance(name, cls):
            name = name.name
        else:
            name = name.upper()
        return name

    def __int__(self):
        return self.value

    def __eq__(self, other):
        try:
            if isinstance(other, int):
                ret = other == self.value

            elif isinstance(other, basestring):
                ret = other == self.name

            else:
                ret = other.value == self.value

        except (AttributeError, TypeError):
            ret = False

        return ret


if is_py2:
    class EnumMeta(type):
        """This is the __metaclass__ for the Enum class, it is basically private to
        this module (and the Enum class). It exists so the Enum class can basically
        act like the built-in python3 enum class in python2

        https://docs.python.org/3/library/enum.html
        """
        def __new__(cls, name, bases, properties):
            """https://docs.python.org/3/library/functions.html#type"""
            instance = super(EnumMeta, cls).__new__(cls, name, bases, properties)

            # if the name of the class isn't Enum then we have a child and we should
            # do some magic
            if name != "Enum":
                members = {}

                for k, v in properties.items():
                    if k.isupper():
                        en = instance(v)
                        en.name = String(k)
                        members[k] = en
                        setattr(instance, k, en)

                # https://docs.python.org/3/library/enum.html#iteration
                instance.__members__ = members
            return instance

        def __contains__(cls, k):
            if isinstance(k, cls):
                k = k.name
            return k in cls.__dict__

        def __getitem__(cls, k):
            if k in cls.__dict__:
                return cls.__dict__[k]

            else:
                raise AttributeError(k)

        def __setattr__(cls, k, v):
            if hasattr(cls, "__members__"):
                # once the __members__ property is set it means we are done adding
                # to this enum instance and it is now read only
                pass
            else:
                super(EnumMeta, cls).__setattr__(k, v)

        def __iter__(cls):
            for v in cls.__members__.values():
                yield v


    class Enum(EnumMixin):
        """Enum Class

        This does its very best to act like the builtin python 3 enum class:

            https://docs.python.org/3/library/enum.html

        Differences from builtin py3 enum class:

            1. This expects enum keys to be uppercase
            2. Enum values can only be integers
            3. This has additional methods that py3 Enum does not have
        """
        __metaclass__ = EnumMeta

        def __init__(self, value):
            self.value = value

else:
    from enum import Enum as BaseEnum, EnumMeta as BaseEnumMeta

    class EnumMeta(BaseEnumMeta):
        def __contains__(cls, k):
            if isinstance(k, cls):
                return super(EnumMeta, cls).__contains__(k)
            else:
                # "FOO" in Enum is deprecated behavior in py3.8, this restores it
                return k in cls.__dict__

    # python 2 parser will fail on metaclass=... syntax, so work around that
    #
    # Order matters for the parent classes
    # https://docs.python.org/3/library/enum.html#restricted-enum-subclassing
    exec("class Enum(EnumMixin, BaseEnum, metaclass=EnumMeta): pass")




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import


def cbany(callback, iterable):
    """Return True if any callback(row) of the iterable is true. If the iterable
    is empty, return False

    :param callback: callable
    :param iterable: Sequence
    :returns: bool
    """
    for v in iterable:
        if callback(v):
            return True
    return False


def cball(callback, iterable):
    """Return True if all elements of the iterable are true (or if the iterable
    is empty)

    :param callback: callable
    :param iterable: Sequence
    :returns: bool
    """
    for v in iterable:
        if not callback(v):
            return False
    return True

